from abc import abstractmethod
from typing import Callable, Dict, List, Tuple

from numpy import sqrt, diag, average
from scipy.optimize import curve_fit

from energyplus_pet.equipment.equip_types import EquipType
from energyplus_pet.equipment.column_header import ColumnHeaderArray
from energyplus_pet.units import UnitType


class BaseEquipment:
    """
    This class represents an abstract piece of equipment to be processed by this library.
    This is primarily a set of abstract methods that define what must be overridden by derived equipment classes.
    Data stored in the equipment class should assume that everything coming from the catalog/correction/rated forms
    are all coming in with the proper calculation units.  The forms won't allow continuing until everything is
    conformed properly.  So equipment specific parameters should be plain float arrays and scalars, not Unit instances.
    """

    def this_type(self) -> EquipType:
        """
        Returns the EquipType enumeration for the derived equipment class.

        :return: Entry from the EquipType enumeration
        """
        return EquipType.InvalidType

    @abstractmethod
    def name(self) -> str:  # pragma: no cover
        """
        Must be overridden to return a meaningful descriptive name for this type of equipment

        :return: String long name for this equipment
        """
        pass

    @abstractmethod
    def short_name(self) -> str:  # pragma: no cover
        """
        Must be overridden to return a brief (max 32 characters) descriptive ID or short name for this type of equipment

        :return: String short name for this equipment
        """
        pass

    @abstractmethod
    def headers(self) -> ColumnHeaderArray:  # pragma: no cover
        """
        Must be overridden to return a set of ColumnHeader instances that define the catalog data for this type of
        equipment

        :return: List of ColumnHeader instances for this equipment
        """
        pass

    class RequiredConstantParameter:
        """A minimal class for capturing information to describe fixed/rated/constant parameters"""
        def __init__(self, p_id: str, title: str, description: str, unit_type: UnitType, default_value: float = 0.0):
            """
            Constructor for the instance

            :param p_id: A unique (for this equipment type) string ID for this parameter
            :param title: A brief (short sentence?) title for this variable
            :param description: A longer (long sentence?) description for this variable
            :param unit_type: A UnitType enum instance for this parameter
            :param default_value: An optional default value, should be already in the UnitType's calculation unit
            """
            self.id = p_id
            self.title = title
            self.description = description
            self.unit_type = unit_type
            self.default_value = default_value

    @abstractmethod
    def get_required_constant_parameters(self) -> List[RequiredConstantParameter]:  # pragma: no cover
        """
        Must be overridden to return a set of constant/fixed/rated parameters for this type of equipment

        :return: List of RequiredConstantParameter instances
        """
        pass

    @abstractmethod
    def set_required_constant_parameter(self, parameter_id: str, new_value: float) -> None:  # pragma: no cover
        """
        Must be overridden to set a single parameter on this equipment instance, accessing it by ID.

        :param parameter_id: ID of the parameter to set, as retrieved from `required_constant_parameters`
        :param new_value: The new floating point value to set as the fixed parameter value, should be in proper units
        :return: Nothing
        """
        pass

    @abstractmethod
    def to_eplus_idf_object(self) -> str:  # pragma: no cover
        """
        Must be overridden to return a fully filled out EnergyPlus IDF object for this type of equipment

        :return: String IDF representation
        """
        pass

    @abstractmethod
    def to_parameter_summary(self) -> str:  # pragma: no cover
        """
        Must be overridden to return a fully filled out free form parameter summary for this type of equipment

        :return: String parameter summary representation
        """
        pass

    @abstractmethod
    def to_eplus_epjson_object(self) -> str:  # pragma: no cover
        """
        Must be overridden to return a fully filled out EnergyPlus EpJSON object for this type of equipment

        :return: String EpJSON representation
        """
        pass

    @abstractmethod
    def get_number_of_progress_steps(self) -> int:  # pragma: no cover
        """
        Must be overridden to return the number of progress increment steps expected during the parameter gen process

        :return: Integer number of progress increment calls expected during processing
        """
        pass

    @abstractmethod
    def minimum_data_points_for_generation(self) -> int:  # pragma: no cover
        """
        Must be overridden to return the minimum number of data points required for generation.
        This is nominally the same as the number of coefficients to be found during curve fitting, but could be set
        higher if more would be necessary.

        :return: Integer number of required data points
        """
        pass

    @abstractmethod
    def generate_parameters(
            self, data_manager, cb_progress_increment: Callable, cb_progress_done: Callable
    ):  # pragma: no cover
        """
        Must be overridden to do the actual processing of parameters.  Most of these will follow a similar pattern of
        taking the generic catalog data manager's final data set, creating meaningful data arrays in this equipment
        instance, using SciPy to perform curve fitting, and finally calculating predicted model outputs from the
        generated coefficients.

        :param data_manager: A fully filled out catalog data manager instance
        :param cb_progress_increment: A callback function to alert the calling form/thread to increment progress.
                                      This callback should not take any extra arguments.
        :param cb_progress_done: A callback function to alert the calling form/thread that the process is complete.
                                 This callback should accept a boolean success flag and a string error message as args.
        :return: Nothing
        """
        pass

    @abstractmethod
    def get_absolute_plot_data(self) -> Tuple:  # pragma: no cover
        """
        Must be overridden to return a tuple of plot data for the raw catalog data comparison plot.  Typically, this
        will be pairs of plot data with catalog data as points and matching model output as lines of the same color.

        :return: A tuple of inner tuples that contain (string data name, string type either 'line' or 'point', a
                 matplotlib color string such as 'red' or 'blue', and a 1D iterable of data points to plot)
        """
        pass

    @abstractmethod
    def get_error_plot_data(self) -> Tuple:  # pragma: no cover
        """
        Must be overridden to return a tuple of plot data for the % error catalog data comparison plot.  Typically, this
        will be a percent error between catalog and model data for the same dependent variable.

        :return: A tuple of inner tuples that contain (string data name, string type typically 'line', a
                 matplotlib color string such as 'red' or 'blue', and a 1D iterable of data points to plot)
        """
        pass

    # noinspection PyMethodMayBeStatic
    def get_extra_regression_metrics(self) -> Tuple:
        """
        Returns a tuple of regression metrics for the parameter generation process.  By default, this returns an empty
        tuple, but can be overridden to return specifics for each type of equipment.

        :return: A tuple of inner tuples that contain (string metric name/description, floating point value)
        """
        return ()

    @staticmethod
    def fill_eplus_object_format(object_name: str, fields: List[Tuple]) -> str:
        """
        This function takes a list of (field name, parameter) tuples and processes
        them into a nicely formed EnergyPlus IDF object.

        :param object_name: An IDF object name
        :param fields: A list of tuples of the form (field name, field value)
        :return: String IDF object
        """
        form = object_name + ',\n'
        field_names = [x[0] for x in fields]
        field_values = [x[1] for x in fields]
        for i, fn in enumerate(field_names):
            if i == len(field_names) - 1:
                form += "{};{}!-" + fn + '\n'
            else:
                form += "{},{}!-" + fn + '\n'
        preferred_spaces = 16
        hanging_indent_string = " "*4
        pads = []
        string_fields = [str(x) for x in field_values]
        for f in string_fields:
            this_padding = preferred_spaces - len(f)
            this_padding = max(2, this_padding)
            pads.append(" "*this_padding)
        all_tokens = []
        for f, p in zip(string_fields, pads):
            all_tokens.append(hanging_indent_string + f)
            all_tokens.append(p)
        return form.format(*all_tokens)

    @staticmethod
    def do_one_curve_fit(
            eval_function: Callable,
            independent_variable_arrays: Tuple[List[float], ...],
            dependent_variable_array: List[float]
    ) -> Tuple[List[float], float]:
        """
        Performs a curve fit operation given an evaluation function, a tuple of arrays of independent variable data,
        and an array of dependent variable values.

        The functional relationship to be solved is: Y = F(X, A), where:

        * Y is the known original dependent data array, such as an array of catalog heat transfer rates.
        * F is the evaluation function, which can take scalar coefficients & independent variables and calculate a new Y
        * X represents the independent variable data, which are passed into F as a tuple argument for each data point
        * A is a set of coefficients to be solved, which are passed to the function as scalar arguments

        :param eval_function: A function that takes a tuple of independent variable scalars and individual coefficient
                              values as trailing arguments. For example
                              ``def calc_one_output((one_flow_rate, one_inlet_temp), coeff_a, coeff_b) -> float``
                              The function should return the output of the curve evaluation.
        :param independent_variable_arrays: A tuple of arrays of independent variable data, for example
                                            ``(flow_rates_all_data_points, inlet_temp_all_data_points)``
        :param dependent_variable_array: A single array of dependent variable data, for example ``all_outputs: [float]``
        :return: Returns a tuple of two items: first is the actual list of solved parameters, and second is a one-sigma
                 average regression error which can be displayed to describe to the user just how good the curve fit is.
        """
        curve_fit_response = curve_fit(
            eval_function,
            independent_variable_arrays,
            dependent_variable_array
        )
        raw_parameters = curve_fit_response[0]
        calculated_parameters = list(raw_parameters)
        average_err_one_sigma = average(sqrt(diag(curve_fit_response[1])))
        return calculated_parameters, average_err_one_sigma

    @staticmethod
    def eval_curve_at_points(
            eval_function: Callable,
            independent_variable_arrays: Tuple[List[float], ...],
            generated_parameter_array: List[float],
            catalog_output_array: List[float],
    ) -> Tuple[List[float], List[float]]:
        """Evaluates a given curve function at all data points given"""
        num_points = len(catalog_output_array)
        predicated_values = []
        error_values = []
        for i in range(num_points):
            predicated_values.append(eval_function(
                [x[i] for x in independent_variable_arrays],
                *generated_parameter_array
            ))
            error_values.append(
                100.0 * (predicated_values[i] - catalog_output_array[i]) /
                catalog_output_array[i]
            )
        return predicated_values, error_values

    @staticmethod
    def current_eplus_version_object_idf() -> str:
        """Returns a version IDF object string to include in IDF outputs"""
        return BaseEquipment.fill_eplus_object_format('Version', [('Version Identifier', 22.2)])

    @staticmethod
    def current_eplus_version_object_epjson() -> Dict:
        """Returns a version EpJSON object to include in EpJSON outputs"""
        return {'Version': {'Version 1': {'version_identifier': 22.2}}}



class CommonCurves:
    """
    This class is just a collection of static curve evaluation functions.
    Several of the pieces of equipment being parameterized here use the same model
    formulation, and there is no need to redefine the evaluation function in each.
    """

    @staticmethod
    def heat_pump_5_coefficient_curve(x, a, b, c, d, e):
        """
        Evaluates:  Y / Y_rated = A + B*(TLI/TLI_R) + C*(TSI/TSI_R) + D*(VLI/VLI_R) + E*(VSI/VSI_R)
        Where Y would represent any of the desired dependent variables, such as Q or Power

        :param x: tuple of independent variables, (TLI/TLI_R, TSI/TSI_R, VLI/VLI_R, VSI/VSI_R)
        :param a: coefficient A in the above equation
        :param b: coefficient B in the above equation
        :param c: coefficient C in the above equation
        :param d: coefficient D in the above equation
        :param e: coefficient E in the above equation
        :return: Scaled dependent variable, such as Q/Q_rated
        """
        return a + b * x[0] + c * x[1] + d * x[2] + e * x[3]

    @staticmethod
    def heat_pump_5_coefficient_curve_raw_value(x, a, b, c, d, e, scale):
        return scale * CommonCurves.heat_pump_5_coefficient_curve(x, a, b, c, d, e)

    @staticmethod
    def heat_pump_6_coefficient_curve(x, a, b, c, d, e, f):
        """
        Evaluates:  Y / Y_rated = A + B*(Tdb/Tdb_R) + C*(Twb/Twb_R) + D*(TSI/TSI_R) + E*(VLI/VLI_R) + F*(VSI/VSI_R)
        Where Y would represent any of the desired dependent variables, such as Q or Power

        :param x: tuple of independent variables, (Tdb/Tdb_R, Twb/Twb_R, TSI/TSI_R, VLI/VLI_R, VSI/VSI_R)
        :param a: coefficient A in the above equation
        :param b: coefficient B in the above equation
        :param c: coefficient C in the above equation
        :param d: coefficient D in the above equation
        :param e: coefficient E in the above equation
        :param f: coefficient F in the above equation
        :return: Scaled dependent variable, such as Q/Q_rated
        """
        return a + b * x[0] + c * x[1] + d * x[2] + e * x[3] + f * x[4]

    @staticmethod
    def heat_pump_6_coefficient_curve_raw_value(x, a, b, c, d, e, f, scale):
        return scale * CommonCurves.heat_pump_6_coefficient_curve(x, a, b, c, d, e, f)



from json import dumps
from typing import Callable, List, Tuple

from energyplus_pet.equipment.common_curves import CommonCurves
from energyplus_pet.data_manager import CatalogDataManager
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.equipment.equip_types import EquipType
from energyplus_pet.equipment.column_header import ColumnHeaderArray, ColumnHeader
from energyplus_pet.exceptions import EnergyPlusPetException
from energyplus_pet.units import UnitType


class WaterToWaterHeatPumpCoolingCurveFit(BaseEquipment):

    def __init__(self):
        # need some rated parameters that we get from the user for scaling, reporting, etc.
        self.rated_load_volume_flow_key = 'vl'
        self.rated_load_volume_flow = 0.0
        self.rated_source_volume_flow_key = 'vs'
        self.rated_source_volume_flow = 0.0
        self.rated_total_capacity_key = 'qc'
        self.rated_total_capacity = 0.0
        self.rated_cooling_power_key = 'cp'
        self.rated_cooling_power = 0.0
        # store some individual arrays for each of the dependent variable input columns
        self.catalog_total_capacity = []
        self.catalog_cooling_power = []
        # these are matrices in the original code, here just store the final vector
        self.total_capacity_params = []
        self.cooling_power_params = []
        # these represent a metric for the quality of the regression
        self.total_capacity_avg_err = 0.0
        self.cooling_power_avg_err = 0.0
        # store the predicted outputs that are calculated from the generated parameters
        self.predicted_total_capacity = []
        self.predicted_cooling_power = []
        self.percent_error_total_capacity = []
        self.percent_error_cooling_power = []
        # store the headers as an instance variable, so we don't recreate on each call to headers()
        self._headers = ColumnHeaderArray(
            [
                ColumnHeader("Source-side Entering Temp", UnitType.Temperature),
                ColumnHeader("Source-side Volume Flow", UnitType.Flow),
                ColumnHeader("Load-side Entering Temp", UnitType.Temperature),
                ColumnHeader("Load-side Volume Flow", UnitType.Flow),
                ColumnHeader("Total Cooling Capacity", UnitType.Power),
                ColumnHeader("Cooling Power", UnitType.Power),
            ]
        )

    def this_type(self) -> EquipType:
        return EquipType.WWHP_Cooling_CurveFit

    def name(self) -> str:
        return "Water to Water Heat Pump, Cooling Coil, Curve Fit Formulation"

    def short_name(self) -> str:
        return "WWHP-Cooling-CurveFit"

    def get_required_constant_parameters(self) -> List[BaseEquipment.RequiredConstantParameter]:
        return [
            # need some rated parameters that we get from the user for scaling, reporting, etc.
            BaseEquipment.RequiredConstantParameter(
                self.rated_load_volume_flow_key,
                "Rated Load-side Flow Rate",
                "This is a nominal flow rate value for the load-side of the coil",
                UnitType.Flow,
                0.0006887,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_source_volume_flow_key,
                "Rated Source-side Flow Rate",
                "This is a nominal flow rate value for the source-side of the coil",
                UnitType.Flow,
                0.0001892,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_total_capacity_key,
                "Rated Total Cooling Capacity",
                "This is a nominal value of the total cooling capacity of the coil",
                UnitType.Power,
                3.513,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_cooling_power_key,
                "Rated Cooling Power",
                "This is a nominal value of the cooling power for this coil",
                UnitType.Power,
                0.900,
            )
        ]

    def set_required_constant_parameter(self, parameter_id: str, new_value: float) -> None:
        if parameter_id == self.rated_load_volume_flow_key:
            self.rated_load_volume_flow = new_value
        elif parameter_id == self.rated_source_volume_flow_key:
            self.rated_source_volume_flow = new_value
        elif parameter_id == self.rated_total_capacity_key:
            self.rated_total_capacity = new_value
        elif parameter_id == self.rated_cooling_power_key:
            self.rated_cooling_power = new_value
        else:
            raise EnergyPlusPetException("Bad parameter ID in set_required_constant_parameter")

    def headers(self) -> ColumnHeaderArray:
        return self._headers

    def to_eplus_idf_object(self) -> str:
        object_name = "HeatPump:WaterToWater:EquationFit:Cooling"
        fields = [
            ("Name", 'Your Cooling Coil Name'),
            ("Source Side Inlet Node Name", 'Your Coil Source Side Inlet Node'),
            ("Source Side Outlet Node Name", 'Your Coil Source Side Outlet Node'),
            ("Load Side Inlet Node Name", 'Your Coil Load Side Inlet Node'),
            ("Load Side Outlet Node Name", 'Your Coil Load Side Outlet Node'),
            ("Reference Load Side Flow Rate", self.rated_load_volume_flow),
            ("Reference Source Side Flow Rate", self.rated_source_volume_flow),
            ("Reference Cooling Capacity", self.rated_total_capacity),
            ("Reference Cooling Power Consumption", self.rated_cooling_power),
            ("Cooling Capacity Curve Name", 'TotalCapacityCurve'),
            ("Cooling Compressor Power Consumption Curve Name", 'CoolingPowerCurve'),
            ("Reference Coefficient of Performance", round(self.rated_total_capacity / self.rated_cooling_power, 8)),
            ("Sizing Factor", ''),
            ("Companion Heating Heat Pump Name", 'Your Heating Coil Name')
        ]
        coil_object_string = self.fill_eplus_object_format(object_name, fields)

        reused_limits = [
            ("Minimum Value of w", -100), ("Maximum Value of w", 100),
            ("Minimum Value of x", -100), ("Maximum Value of x", 100),
            ("Minimum Value of y", -100), ("Maximum Value of y", 100),
            ("Minimum Value of z", -100), ("Maximum Value of z", 100),
        ]

        object_name = "Curve:QuadLinear"
        fields = [
            ("Name", "TotalCapacityCurve"),
            *[(f"Coefficient{i}", self.total_capacity_params[i]) for i in range(5)],
            *reused_limits
        ]
        total_curve_output = self.fill_eplus_object_format(object_name, fields)

        object_name = "Curve:QuadLinear"
        fields = [
            ("Name", "CoolingPowerCurve"),
            *[(f"Coefficient{i}", self.cooling_power_params[i]) for i in range(5)],
            *reused_limits
        ]
        power_curve_output = self.fill_eplus_object_format(object_name, fields)

        return '\n'.join([
            BaseEquipment.current_eplus_version_object_idf(),
            coil_object_string,
            total_curve_output, power_curve_output
        ])

    def to_parameter_summary(self) -> str:
        output = f"""{self.name()}
**Begin Nomenclature**
TC: Total Cooling Capacity
CP: Cooling Power Consumption
TLI: Entering Load-side Temperature
TSI: Entering Source-side Temperature
VLI: Entering Load-side Flow Rate
VSI: Entering Source-side Flow Rate
Subscript _R: Rated Value
Subscript _#: Coefficient #
**End Nomenclature**

**Begin Governing Equations**
(TC/TC_R) = TC_1 + TC_2*(TLI/TLI_R) + TC_3*(TSI/TSI_R) + TC_4*(VLI/VLI_R) + TC_5*(VSI/VSI_R)
(CP/CP_R) = CP_1 + CP_2*(TLI/TLI_R) + CP_3*(TSI/TSI_R) + CP_4*(VLI/VLI_R) + CP_5*(VSI/VSI_R)
**End Governing Equations**

**Begin Reporting Parameters**
Rated Load-side Total Cooling Capacity: {self.rated_total_capacity}
Rated Cooling Power Consumption: {self.rated_cooling_power}
Rated Load-side Volumetric Flow Rate: {self.rated_load_volume_flow}
Rated Source-side Volumetric Flow Rate: {self.rated_source_volume_flow}
"""
        for i, c in enumerate(self.total_capacity_params):
            output += f"Cooling Total Capacity Coefficient TC_{i + 1}: {round(c, 4)}\n"
        for i, c in enumerate(self.cooling_power_params):
            output += f"Cooling Power Consumption Coefficient CP_{i + 1}: {round(c, 4)}\n"
        output += "**End Reporting Parameters**"
        return output

    def to_eplus_epjson_object(self) -> str:
        coil_object = {'Your Cooling Coil Name': {
            'source_side_inlet_node_name': 'Your Coil Source Side Inlet Node',
            'source_side_outlet_node_name': 'Your Coil Source Side Outlet Node',
            'load_side_inlet_node_name': 'Your Coil Load Side Inlet Node',
            'load_side_outlet_node_name': 'Your Coil Load Side Outlet Node',
            'reference_load_side_flow_rate': self.rated_load_volume_flow,
            'reference_source_side_flow_rate': self.rated_source_volume_flow,
            'reference_cooling_capacity': self.rated_total_capacity,
            'reference_cooling_power_consumption': self.rated_cooling_power,
            'cooling_capacity_curve_name': 'TotalCapacityCurve',
            'cooling_compressor_power_curve_name': 'CoolingPowerCurve',
            'reference_coefficient_of_performance': self.rated_total_capacity / self.rated_cooling_power,
            'sizing_factor': '',
            'companion_heating_heat_pump_name': 'Your Heating Coil Name',
        }}
        reused_limits = {
            "minimum_value_of_w": -100, "maximum_value_of_w": 100,
            "minimum_value_of_x": -100, "maximum_value_of_x": 100,
            "minimum_value_of_y": -100, "maximum_value_of_y": 100,
            "minimum_value_of_z": -100, "maximum_value_of_z": 100,
        }
        curves = {'TotalCapacityCurve': {
            "coefficient1_constant": self.total_capacity_params[0],
            "coefficient2_w": self.total_capacity_params[1],
            "coefficient3_x": self.total_capacity_params[2],
            "coefficient4_y": self.total_capacity_params[3],
            "coefficient5_z": self.total_capacity_params[4],
            **reused_limits
        }, 'CoolingPowerCurve': {
            "coefficient1_constant": self.cooling_power_params[0],
            "coefficient2_w": self.cooling_power_params[1],
            "coefficient3_x": self.cooling_power_params[2],
            "coefficient4_y": self.cooling_power_params[3],
            "coefficient5_z": self.cooling_power_params[4],
            **reused_limits
        }}

        epjson_object = {
            **BaseEquipment.current_eplus_version_object_epjson(),
            'HeatPump:WaterToWater:EquationFit:Cooling': coil_object,
            'Curve:QuadLinear': curves
        }
        return dumps(epjson_object, indent=2)

    def get_number_of_progress_steps(self) -> int:
        return 3  # read data, tc curve fit, power curve fit

    def minimum_data_points_for_generation(self) -> int:
        return 5

    def generate_parameters(
            self, data_manager: CatalogDataManager, cb_progress_increment: Callable, cb_progress_done: Callable
    ):
        # step 1, read the data into arrays (will be used for both scaling calculations and plotting later)
        # these should already be set to zero, but not sure if the main form will reinitialize the equip-instance or not
        self.catalog_total_capacity = [x[4] for x in data_manager.final_data_matrix]
        self.catalog_cooling_power = [x[5] for x in data_manager.final_data_matrix]
        scaled_source_side_inlet_temp = []
        scaled_source_side_flow_rate = []
        scaled_load_side_inlet_temp = []
        scaled_load_side_flow_rate = []
        scaled_cooling_capacity = []
        scaled_cooling_power = []
        for data_row in data_manager.final_data_matrix:
            scaled_source_side_inlet_temp.append((data_row[0] + 273.15) / (10.0 + 273.15))  # T_ref defined by HP model
            scaled_source_side_flow_rate.append(data_row[1] / self.rated_source_volume_flow)
            scaled_load_side_inlet_temp.append((data_row[2] + 273.15) / (10.0 + 273.15))
            scaled_load_side_flow_rate.append(data_row[3] / self.rated_load_volume_flow)
            scaled_cooling_capacity.append(data_row[4] / self.rated_total_capacity)
            scaled_cooling_power.append(data_row[5] / self.rated_cooling_power)
        cb_progress_increment()

        independent_var_arrays = (
            scaled_load_side_inlet_temp,
            scaled_source_side_inlet_temp,
            scaled_load_side_flow_rate,
            scaled_source_side_flow_rate
        )

        self.total_capacity_params, self.total_capacity_avg_err = self.do_one_curve_fit(
            CommonCurves.heat_pump_5_coefficient_curve,
            independent_var_arrays,
            scaled_cooling_capacity
        )
        cb_progress_increment()

        self.cooling_power_params, self.cooling_power_avg_err = self.do_one_curve_fit(
            CommonCurves.heat_pump_5_coefficient_curve,
            independent_var_arrays,
            scaled_cooling_power
        )
        cb_progress_increment()

        # now just recalculate the values at each catalog data point
        self.predicted_total_capacity, self.percent_error_total_capacity = self.eval_curve_at_points(
            lambda *x: CommonCurves.heat_pump_5_coefficient_curve_raw_value(*x, self.rated_total_capacity),
            independent_var_arrays,
            self.total_capacity_params,
            self.catalog_total_capacity
        )
        self.predicted_cooling_power, self.percent_error_cooling_power = self.eval_curve_at_points(
            lambda *x: CommonCurves.heat_pump_5_coefficient_curve_raw_value(*x, self.rated_cooling_power),
            independent_var_arrays,
            self.cooling_power_params,
            self.catalog_cooling_power
        )
        cb_progress_done(True)

    def get_absolute_plot_data(self) -> Tuple:
        return (
            ('Total Heat Transfer Model', 'line', 'red', self.predicted_total_capacity),
            ('Total Heat Transfer Catalog', 'point', 'red', self.catalog_total_capacity),
            ('Cooling Power Model', 'line', 'green', self.predicted_cooling_power),
            ('Cooling Power Catalog', 'point', 'green', self.catalog_cooling_power),
        )

    def get_error_plot_data(self) -> Tuple:
        return (
            ('Total Heat Transfer % Error', 'line', 'red', self.percent_error_total_capacity),
            ('Cooling Power % Error', 'line', 'green', self.percent_error_cooling_power),
        )

    def get_extra_regression_metrics(self) -> Tuple:
        return (
            (
                "Total Heat Transfer Average curve-fit error (1 standard deviation)",
                self.total_capacity_avg_err
            ),
            (
                "Cooling Power Average curve-fit error (1 standard deviation)",
                self.cooling_power_avg_err
            )
        )



from enum import Enum, auto
from tkinter import Toplevel, Frame  # containers
from tkinter import Button, Label, OptionMenu  # widgets
from tkinter import HORIZONTAL, TOP, X, W, EW, BOTH  # appearance attributes
from tkinter import StringVar  # dynamic variables
from tkinter.ttk import Separator

from tksheet import Sheet

from energyplus_pet.correction_factor import CorrectionFactor, CorrectionFactorType
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.forms.basic_message_form import PetMessageForm
from energyplus_pet.exceptions import EnergyPlusPetException
from energyplus_pet.units import unit_class_factory, unit_instance_factory, TemperatureValue, UnitType


class DetailedCorrectionFactorForm(Toplevel):
    """
    This form is where the user will enter correction factor detailed tabular data
    """

    class DetailedCorrectionExitCode(Enum):
        Done = auto()
        Cancel = auto()

    def __init__(self, parent_window, _cf: CorrectionFactor, eq: BaseEquipment, cf_num: int, num_cfs: int):
        super().__init__(parent_window, height=200, width=200)
        p = 4
        self.completed_factor = _cf
        self.equipment_instance = eq
        self.exit_code = DetailedCorrectionFactorForm.DetailedCorrectionExitCode.Cancel
        self.need_to_conform_units = False
        # create all the objects
        self.title(f"{parent_window.title()}: Enter Correction Factor Data {cf_num}/{num_cfs}")
        base_column_name = eq.headers().name_array()[_cf.base_column_index]
        dep_column_names = ""
        for mod_column in _cf.columns_to_modify:
            dep_column_names += f"  -- {eq.headers().name_array()[mod_column]}\n"

        #
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
        #
        if _cf.correction_type == CorrectionFactorType.CombinedDbWb:
            Label(self, text=f"""Entering correction data for ~~ {_cf.name} ~~

This correction factor requires {_cf.num_corrections} replacement values for both dry and wet bulb temps.

The correction factor requires multiplier values for the following {len(_cf.columns_to_modify)} column(s):
{dep_column_names}""").pack(side=TOP, fill=X, anchor=W, expand=False, padx=p, pady=p)
            temp_string_options = list(TemperatureValue.get_unit_string_map().values())
            self.preferred_db_wb_unit_id = TemperatureValue.calculation_unit_id()
            self.preferred_db_wb_unit_string = TemperatureValue.get_unit_string_map()[self.preferred_db_wb_unit_id]
            self._tk_var_db_units_string = StringVar(value=self.preferred_db_wb_unit_string)
            self._tk_var_wb_units_string = StringVar(value=self.preferred_db_wb_unit_string)
            self._tk_var_db_units_string.trace('w', self._units_changed)
            self._tk_var_wb_units_string.trace('w', self._units_changed)
            db_wb_frame = Frame(self)
            db_unit_string = f"Units for DB Column: \"{eq.headers().name_array()[eq.headers().get_db_column()]}\""
            Label(db_wb_frame, text=db_unit_string).grid(row=0, column=0, padx=p, pady=p)
            OptionMenu(db_wb_frame, self._tk_var_db_units_string, *temp_string_options).grid(
                row=0, column=1, padx=p, pady=p
            )
            wb_unit_string = f"Units for WB Column: \"{eq.headers().name_array()[eq.headers().get_wb_column()]}\""
            Label(db_wb_frame, text=wb_unit_string).grid(row=1, column=0, padx=p, pady=p)
            OptionMenu(db_wb_frame, self._tk_var_wb_units_string, *temp_string_options).grid(
                row=1, column=1, padx=p, pady=p
            )
            db_wb_frame.pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
            Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)

        elif _cf.correction_type == CorrectionFactorType.Replacement:
            Label(self, text=f"""Entering correction data for ~~ {_cf.name} ~~

This correction factor requires {_cf.num_corrections} replacement values for both dry and wet bulb temps.

The correction factor requires multiplier values for the following {len(_cf.columns_to_modify)} column(s):
{dep_column_names}""").pack(side=TOP, fill=X, anchor=W, expand=False, padx=p, pady=p)
            replacement_frame = Frame(self)
            replacement_label = Label(replacement_frame, text="Specify the units of the replacement data (column 1):")
            # all we need to do is look up the UnitType for the replacement column
            # from this type, we can get a Base unit class and get everything we need from it.
            # we don't need to create unit-value instances until we are conforming the final data
            # this returns a UnitType enum type of unit, such as UnitType.Temperature or UnitType.FlowRate
            self.replacement_column_unit_type = eq.headers().unit_array()[_cf.base_column_index]
            # this returns a class "type" not an instance, so like TemperatureUnits or PowerUnits
            self.unit_type_class = unit_class_factory(self.replacement_column_unit_type)
            # for this specific unit type, get a mapping of the underlying ids to user-consumable names {'some_id': 'W'}
            self.unit_id_to_string_mapping = self.unit_type_class.get_unit_string_map()
            # one thing we need is a list of all the user-facing unit names, such as ['W', 'kW']
            replacement_unit_strings = list(self.unit_id_to_string_mapping.values())
            # store the preferred replacement_unit id and string
            preferred_replacement_unit_id = self.unit_type_class.calculation_unit_id()
            self.preferred_replacement_unit_string = self.unit_id_to_string_mapping[preferred_replacement_unit_id]
            # create a Tk variable to store the currently shown user string for replacement units
            self._tk_var_replacement_units_string = StringVar(value=self.preferred_replacement_unit_string)
            self._tk_var_replacement_units_string.trace('w', self._units_changed)
            replacement_option = OptionMenu(
                replacement_frame, self._tk_var_replacement_units_string, *replacement_unit_strings
            )
            replacement_label.grid(row=0, column=0, padx=p, pady=p)
            replacement_option.grid(row=0, column=1, sticky=EW, padx=p, pady=p)
            replacement_frame.pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
            Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)

        else:  # regular multiplier
            Label(self, text=f"""Entering correction data for ~~ {_cf.name} ~~

This correction factor requires {_cf.num_corrections} multiplier values for {base_column_name}.

The correction factor requires multiplier values for the following {len(_cf.columns_to_modify)} column(s):
{dep_column_names}""").pack(side=TOP, fill=X, anchor=W, expand=False, padx=p, pady=p)

        self.tabular_frame = Frame(self)
        column_titles = list()
        # build out the independent variables first
        if self.completed_factor.correction_type != CorrectionFactorType.CombinedDbWb:
            # just add the one selected base column
            column_titles.append(eq.headers().name_array()[_cf.base_column_index])
        else:
            # this is a db/wb entry, need to add db to column 1 and wb to column 2
            column_titles.append(eq.headers().name_array()[eq.headers().get_db_column()])
            column_titles.append(eq.headers().name_array()[eq.headers().get_wb_column()])
        # then add the dependent variables
        for mod_column in _cf.columns_to_modify:
            column_titles.append(eq.headers().name_array()[mod_column])
        num_columns_in_table = len(column_titles)
        num_rows_in_table = self.completed_factor.num_corrections  # no need to add a header or anything here
        pretend_data = []
        for row in range(num_rows_in_table):
            this_row = []
            for col in range(num_columns_in_table):
                this_row.append(0.0)
            pretend_data.append(this_row)
        self._num_bad_cells = 0
        self.table = Sheet(self.tabular_frame)
        self.table.set_sheet_data(pretend_data)
        self.table.headers(column_titles)
        self.table.enable_bindings()
        self.table.set_options(expand_sheet_if_paste_too_big=False)
        self.table.hide(canvas="row_index")
        self.table.hide(canvas="top_left")
        self.table.set_all_cell_sizes_to_text(redraw=True)
        # https://github.com/ragardner/tksheet/blob/master/DOCUMENTATION.md#25-example-custom-right-click-and-text-editor-functionality
        self.table.extra_bindings("end_edit_cell", func=self._cell_edited)
        self.table.extra_bindings("end_paste", func=self._cells_pasted)
        self.table.pack(side=TOP, expand=True, fill=BOTH, padx=p, pady=p)
        self.table.select_cell(row=0, column=0)
        self.table.focus()
        self.tabular_frame.pack(side=TOP, fill=BOTH, expand=True, padx=p, pady=p)
        # we block resizing the table when pasting, and we are hiding the row titles to block inserting rows via title,
        # but we can't seem to block tksheet from allowing a column insert, so we will just have to check later
        # https://github.com/ragardner/tksheet/blob/0ed6ee8611a7062eedfe43333e7760e1f22088af/tksheet/_tksheet_main_table.py#L3172
        # store the expected number of columns and if it doesn't match later just make the user fix it
        self.expected_num_columns: int = self.table.total_columns()
        #
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
        #
        button_frame = Frame(self)
        self._tk_var_done_conform_text = StringVar(value="This one is done, continue")
        btn_done_conform = Button(
            button_frame, textvariable=self._tk_var_done_conform_text, command=self._done_or_conform
        )
        btn_cancel = Button(button_frame, text="Cancel Wizard", command=self.cancel)
        btn_done_conform.grid(row=4, column=0, padx=p, pady=p)
        btn_cancel.grid(row=4, column=1, padx=p, pady=p)
        button_frame.grid_columnconfigure(0, weight=1)
        button_frame.grid_columnconfigure(1, weight=1)
        button_frame.pack(side=TOP, fill=X, expand=False, padx=p, pady=p)

        # draw factors, in case there already are any
        # self.draw_factors()

        # set up keybinding and flags
        self.bind('<Key>', self._handle_button_pressed)

        # finalize UI operations
        self.grab_set()
        self.transient(parent_window)

    def _handle_button_pressed(self, event):
        # relevant_modifiers
        # mod_shift = 0x1
        mod_control = 0x4
        # mod_alt = 0x20000
        if event.keysym == 'd' and mod_control & event.state:
            for row in range(self.completed_factor.num_corrections):
                for col in range(len(self.completed_factor.columns_to_modify) + 1):
                    self.table.set_cell_data(row, col, (row + 1.0) * (col + 1.0))
            self._cells_pasted()

    def _cell_edited(self, event):
        row = event.row
        column = event.column
        # first check the current value to see if it was already bad -- it will be the old value in the cell
        current_value = self.table.get_cell_data(row, column)
        try:
            float(current_value)
            was_good = True
        except ValueError:
            was_good = False
        try:
            float(event.text)
            self.table.highlight_cells(row, column, bg='white')
            # if it was good already, and good now, we are done
            # if it was bad, and now good, decrement the counter
            if not was_good:
                self._num_bad_cells -= 1
        except ValueError:
            self.table.highlight_cells(row, column, bg='pink')
            # if it was bad already, and bad now, we are done
            # if it was good, and now bad, increment the counter
            if was_good:
                self._num_bad_cells += 1

    def _cells_pasted(self, _=None):
        self._num_bad_cells = 0
        for row in range(self.table.total_rows()):
            for column in range(self.table.total_columns()):
                try:
                    float(self.table.get_cell_data(row, column))
                    self.table.highlight_cells(row, column, bg='white')
                except ValueError:
                    self._num_bad_cells += 1
                    self.table.highlight_cells(row, column, bg='pink')

    def _table_data_all_good(self) -> bool:
        return self._num_bad_cells == 0

    def any_blank_cells(self):
        for row in range(self.table.total_rows()):
            for col in range(self.table.total_columns()):
                if self.table.get_cell_data(row, col) == '':
                    return True
        return False

    def _units_changed(self, *_):
        """This function is called back by the unit traces, just check all of them"""
        self.need_to_conform_units = False
        if self.completed_factor.correction_type == CorrectionFactorType.Multiplier:
            pass  # nothing to be done, there are no units to conform
        elif self.completed_factor.correction_type == CorrectionFactorType.Replacement:
            if self._tk_var_replacement_units_string.get() != self.preferred_replacement_unit_string:
                self.need_to_conform_units = True
        elif self.completed_factor.correction_type == CorrectionFactorType.CombinedDbWb:
            if self._tk_var_db_units_string.get() != self.preferred_db_wb_unit_string:
                self.need_to_conform_units = True
            if self._tk_var_wb_units_string.get() != self.preferred_db_wb_unit_string:
                self.need_to_conform_units = True
        self._refresh_done_conform_button_text()

    def _refresh_done_conform_button_text(self):
        if self.need_to_conform_units:
            self._tk_var_done_conform_text.set("Conform units")
        else:
            self._tk_var_done_conform_text.set("This one is done, continue")
        self.table.redraw()

    def _done_or_conform(self):
        # don't try to conorm or exit if there are any bad values
        if not self._table_data_all_good():
            message_window = PetMessageForm(
                self, "Value Issue", f"{self._num_bad_cells} cell(s) appear to have bad values; fix them and retry"
            )
            self.wait_window(message_window)
            return
        if self.need_to_conform_units:
            self.conform_units()
        else:
            if self.expected_num_columns != self.table.total_columns():
                message_window = PetMessageForm(
                    self, "Problem with Table Size", "Table column size does not match expected size, cancel and retry"
                )
                self.wait_window(message_window)
                return
            if self.any_blank_cells():
                message_window = PetMessageForm(
                    self,
                    "Problem with Table Entries",
                    "Blank entries detected in the table, fix and retry"
                )
                self.wait_window(message_window)
                return
            for r in range(self.table.total_rows()):
                for c in range(self.table.total_columns()):
                    value = self.table.get_cell_data(r, c)
                    if value == '':
                        message_window = PetMessageForm(
                            self,
                            "Problem with Correction Factor Data",
                            "Encountered at least one blank cell in the data, populate the full table and retry"
                        )
                        self.wait_window(message_window)
                        return
            # OK, so if we are done, units must be ready, so we just need to update the data in the correction factor
            # for regular correction factors, read the first column of data into the base correction of the CF
            if self.completed_factor.correction_type != CorrectionFactorType.CombinedDbWb:
                self.completed_factor.base_correction = [
                    float(self.table.get_cell_data(row, 0)) for row in range(self.completed_factor.num_corrections)
                ]
                last_column_read = 0
            else:  # for db/wb, read the first column into the db array and the second column into the wb array
                self.completed_factor.base_correction_db = [
                    float(self.table.get_cell_data(row, 0)) for row in range(self.completed_factor.num_corrections)
                ]
                self.completed_factor.base_correction_wb = [
                    float(self.table.get_cell_data(row, 1)) for row in range(self.completed_factor.num_corrections)
                ]
                last_column_read = 1
            # then iterate over the mod column ids of the CF and read the data into lists and then into the CF map
            for equipment_column_index in self.completed_factor.columns_to_modify:
                last_column_read += 1
                this_column = []
                for row in range(self.table.total_rows()):
                    this_column.append(float(self.table.get_cell_data(row, last_column_read)))
                self.completed_factor.mod_correction_data_column_map[equipment_column_index] = this_column
            output_message = ""
            db = self.equipment_instance.headers().get_db_column()
            wb = self.equipment_instance.headers().get_wb_column()
            if not self.completed_factor.check_ok(db, wb):
                output_message += f"Errors for {self.completed_factor.name}:\n"
                for e in self.completed_factor.check_ok_messages:
                    output_message += f" - {e}\n"
            if output_message:
                message_window = PetMessageForm(
                    self, "Problem with Correction Factor", output_message, justify_message_left=True
                )
                self.wait_window(message_window)
                return
            self.exit_code = DetailedCorrectionFactorForm.DetailedCorrectionExitCode.Done
            self.grab_release()
            self.destroy()

    def conform_units(self):
        if self.completed_factor.correction_type == CorrectionFactorType.Multiplier:
            pass  # shouldn't have been able to get here, just carry on, nothing to do
        elif self.completed_factor.correction_type == CorrectionFactorType.Replacement:
            db_units_string = self._tk_var_replacement_units_string.get()
            try:
                current_db_unit_id = self.unit_type_class.get_id_from_unit_string(db_units_string)
            except EnergyPlusPetException:
                pmf = PetMessageForm(self, "Unit ID Problem", "Could not match unit ID; this is a developer issue.")
                self.wait_window(pmf)
                self.cancel()
                return
            for r in range(self.table.total_rows()):
                cell_value = float(self.table.get_cell_data(r, 0))
                unit_value = unit_instance_factory(cell_value, self.replacement_column_unit_type)
                unit_value.units = current_db_unit_id
                unit_value.convert_to_calculation_unit()
                self.table.set_cell_data(r, 0, unit_value.value)
            self._tk_var_replacement_units_string.set(self.preferred_replacement_unit_string)
        elif self.completed_factor.correction_type == CorrectionFactorType.CombinedDbWb:
            db_units_string = self._tk_var_db_units_string.get()
            try:
                current_db_unit_id = TemperatureValue.get_id_from_unit_string(db_units_string)
            except EnergyPlusPetException:
                pmf = PetMessageForm(self, "Unit ID Problem", "Could not match DB unit ID; this is a developer issue.")
                self.wait_window(pmf)
                self.cancel()
                return
            for r in range(self.table.total_rows()):
                cell_value = float(self.table.get_cell_data(r, 0))
                unit_value = unit_instance_factory(cell_value, UnitType.Temperature)
                unit_value.units = current_db_unit_id
                unit_value.convert_to_calculation_unit()
                self.table.set_cell_data(r, 0, unit_value.value)
            self._tk_var_db_units_string.set(self.preferred_db_wb_unit_string)
            wb_units_string = self._tk_var_wb_units_string.get()
            try:
                current_wb_unit_id = TemperatureValue.get_id_from_unit_string(wb_units_string)
            except EnergyPlusPetException:
                pmf = PetMessageForm(self, "Unit ID Problem", "Could not match WB unit ID; this is a developer issue.")
                self.wait_window(pmf)
                self.cancel()
                return
            for r in range(self.table.total_rows()):
                cell_value = float(self.table.get_cell_data(r, 1))
                unit_value = unit_instance_factory(cell_value, UnitType.Temperature)
                unit_value.units = current_wb_unit_id
                unit_value.convert_to_calculation_unit()
                self.table.set_cell_data(r, 1, unit_value.value)
            self._tk_var_wb_units_string.set(self.preferred_db_wb_unit_string)
        self.need_to_conform_units = False
        self._refresh_done_conform_button_text()

    def cancel(self):
        self.exit_code = DetailedCorrectionFactorForm.DetailedCorrectionExitCode.Cancel
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_cooling_curve import WaterToAirHeatPumpCoolingCurveFit

    root = Tk()
    cf = CorrectionFactor('Load Side DB/WB Temperature Correction')
    cf.num_corrections = 2
    cf.base_column_index = 0
    cf.columns_to_modify = [4, 5, 6]
    cf.correction_type = CorrectionFactorType.Multiplier
    _eq = WaterToAirHeatPumpCoolingCurveFit()
    DetailedCorrectionFactorForm(root, cf, _eq, 1, 1)
    root.mainloop()



from tkinter import Toplevel, Frame, LabelFrame  # containers
from tkinter import Button, Label, OptionMenu, Entry  # widgets
from tkinter import TOP, X, RIDGE, BOTH, ALL, END  # appearance stuff
from tkinter import StringVar, DoubleVar  # dynamic variables
from tkinter import TclError
from typing import Callable

from energyplus_pet.forms.basic_message_form import PetMessageForm
from energyplus_pet.units import unit_class_factory, unit_instance_factory
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.exceptions import EnergyPlusPetException


class ConstantParameterEntryWidget(Frame):
    """
    This form is where the user enters any fixed/rated parameter values for this equipment type
    """
    def __init__(
            self, rp: BaseEquipment.RequiredConstantParameter, parent_frame: LabelFrame, units_ch_handler: Callable
    ):
        """
        Constructor for this widget.  Creates a small grid for users to enter a single parameter.
        Four member variables are public:

        * rp_id (str) which is the unique id for this parameter for this equipment type
        * units_conformed (bool) which is a flag for whether the units are ready to go here
        * var_value (Tk.DoubleVar) which holds the value of the parameter, usually accessed after conforming
        * var_units_string (Tk.StringVar) which holds the units of the parameter, usually accessed after conforming

        :param rp: A required constant/rated parameter, defined on each equipment type
        :param parent_frame: A frame to hold this widget
        :param units_ch_handler: A function to be called back to alert the parent of units changes
        """
        super().__init__(parent_frame, borderwidth=1, relief=RIDGE)
        self.rp_id = rp.id
        self._units_changed_handler = units_ch_handler
        p = 4
        self._unit_type = rp.unit_type
        self._unit_type_class = unit_class_factory(self._unit_type)
        self._unit_id_to_string_mapping = self._unit_type_class.get_unit_string_map()
        unit_strings = list(self._unit_id_to_string_mapping.values())
        preferred_unit_id = self._unit_type_class.calculation_unit_id()
        self._preferred_unit_string = self._unit_id_to_string_mapping[preferred_unit_id]
        Label(self, text=rp.title).grid(row=0, column=0, padx=p, pady=p)
        Label(self, text=rp.description).grid(row=1, column=0, padx=p, pady=p)
        # create a tk variable to hold the value in the entry, don't assign it a value yet
        self._tk_var_value = DoubleVar()
        # create the entry pointing it back to the tk variable
        self.entry = Entry(self, textvariable=self._tk_var_value)
        self.entry.grid(row=0, column=1, padx=p, pady=p)
        self.entry.bind('<FocusIn>', lambda x: self.entry.selection_range(0, END))
        # set up a trace to the value to track any changes and assign it in order to initialize self.valid_number flag
        self._tk_var_value.trace('w', self._validate_number)
        self._tk_var_value.set(rp.default_value)
        self.var_units_string = StringVar(value=self._preferred_unit_string)
        o = OptionMenu(self, self.var_units_string, *unit_strings, command=self._units_changed)
        o.config(takefocus=1)
        o.grid(row=1, column=1, padx=p, pady=p)
        self.grid_columnconfigure(ALL, weight=1)
        self.grid_rowconfigure(ALL, weight=1)
        self.units_conformed = True

    def _units_changed(self, proposed_units):
        """Handler for user changing units, just sets a flag for the parent form"""
        self.units_conformed = proposed_units == self._preferred_unit_string
        self._units_changed_handler()

    def _validate_number(self, *_):
        try:
            float(self._tk_var_value.get())
            self.valid_number = True
            self.entry['background'] = 'white'
        except TclError:
            self.valid_number = False
            self.entry['background'] = 'pink'

    def current_value(self):
        """Returns the private value of the numeric entry"""
        return self._tk_var_value.get()

    def conform_units(self):
        """
        Conforms the value to the preferred units for this unit type

        :return: Nothing
        """
        current_units_string = self.var_units_string.get()
        try:
            current_unit_id = self._unit_type_class.get_id_from_unit_string(current_units_string)
        except EnergyPlusPetException:
            pmf = PetMessageForm(self, "Unit ID Problem", "Could not match unit ID; this is a developer issue.")
            self.wait_window(pmf)
            self.units_conformed = False
            return
        current_value = self._tk_var_value.get()
        unit_value = unit_instance_factory(current_value, self._unit_type)
        unit_value.units = current_unit_id
        unit_value.convert_to_calculation_unit()
        self._tk_var_value.set(unit_value.value)
        self.var_units_string.set(self._preferred_unit_string)
        self.units_conformed = True

    def set_focus_to_entry(self):
        """Sets the focus to the entry box, used for initializing the form"""
        self.entry.focus()


class ConstantParameterEntryForm(Toplevel):

    def __init__(self, parent_window, equipment: BaseEquipment):
        super().__init__(parent_window)
        self.title(f"{parent_window.title()}: Rated/Constant Parameter Entry")
        self.form_cancelled = True
        self.equipment = equipment
        entries_form = LabelFrame(self, text="Required Parameters")
        self.parameter_value_map = {}
        self.known_parameters = []
        for rp in equipment.get_required_constant_parameters():
            c = ConstantParameterEntryWidget(rp, entries_form, self.check_all_units)
            c.pack(side=TOP, fill=X, expand=True, padx=4, pady=4)
            self.known_parameters.append(c)
        entries_form.pack(side=TOP, expand=True, fill=BOTH, padx=4, pady=4)
        self.need_to_conform_units = False  # initially it assumes the calculation unit
        button_frame = Frame(self)
        self.var_conform_done = StringVar(value="Done, process data now")
        Button(button_frame, textvariable=self.var_conform_done, command=self.conform_done).grid(
            row=0, column=0, padx=3, pady=3
        )
        Button(button_frame, text="Cancel Wizard", command=self.close_me).grid(row=0, column=1, padx=3, pady=3)
        button_frame.grid_columnconfigure(0, weight=1)
        button_frame.grid_columnconfigure(1, weight=1)
        button_frame.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        self.check_all_units()
        self.known_parameters[0].set_focus_to_entry()
        # finalize UI operations
        self.grab_set()
        self.transient(parent_window)

    def check_all_units(self):
        self.need_to_conform_units = False
        for parameter_widget in self.known_parameters:
            if not parameter_widget.units_conformed:
                self.var_conform_done.set("Conform units")
                self.need_to_conform_units = True
                return
        self.var_conform_done.set("Done, process data now")

    def conform_done(self):
        if self.need_to_conform_units:
            for parameter_widget in self.known_parameters:
                parameter_widget.conform_units()
            self.check_all_units()
        else:
            if any([not x.valid_number for x in self.known_parameters]):
                pmf = PetMessageForm(self, "Numeric Issue", "At least one entry has an invalid value, fix and retry.")
                self.wait_window(pmf)
                return
            for parameter_widget in self.known_parameters:
                self.parameter_value_map[parameter_widget.rp_id] = parameter_widget.current_value()
            self.form_cancelled = False
            self.close_me()

    def close_me(self):
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_heating_curve import WaterToAirHeatPumpHeatingCurveFit
    tk = Tk()
    e = WaterToAirHeatPumpHeatingCurveFit()
    f = ConstantParameterEntryForm(tk, e)
    tk.mainloop()



from enum import Enum, auto
from tkinter import Toplevel, Frame, LabelFrame  # containers
from tkinter import Button, Label  # widgets
from tkinter import HORIZONTAL, TOP, X, BOTH, ALL  # appearance attributes
from tkinter import StringVar  # dynamic variables
from tkinter.ttk import Separator
from typing import List

from tksheet import Sheet

from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.exceptions import EnergyPlusPetException
from energyplus_pet.units import unit_class_factory, unit_instance_factory
from energyplus_pet.forms.basic_message_form import PetMessageForm


class MainDataForm(Toplevel):
    """
    This form is where the user will enter/paste the raw (typically pretty large) base data set, which will ultimately
    be processed with any correction factors to determine the final data set.
    """
    class MainDataExitCode(Enum):
        Done = auto()
        Cancel = auto()

    def __init__(self, parent_window, eq: BaseEquipment):
        super().__init__(parent_window, height=200, width=200)
        self.title(f"{parent_window.title()}: Main Catalog Data Input")
        p = 4
        self.equipment_instance = eq
        self.exit_code = MainDataForm.MainDataExitCode.Done
        self.need_to_conform_units = False
        self.final_base_data_rows: List[List[float]] = [[]]
        # create all the objects
        Label(
            self, text="""Now it's time to copy in the bulk of the catalog data.
To be brought in properly, the data should be divided into columns, delimited by tab characters.
This is the way most spreadsheets will store data in the clipboard, so the preferred approach is
to paste/cleanup the data in a spreadsheet, then copy the data and paste directly in the table here."""
        ).pack(
            side=TOP, fill=X, expand=False, padx=p, pady=p
        )
        #
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
        #
        tabular_frame = Frame(self)
        column_titles = eq.headers().name_array()
        column_units = eq.headers().unit_array()
        self.table = Sheet(tabular_frame)
        pretend_data = []
        # Would be nice to create a struct of extra data and tag each column with it rather than separate lists
        self.columnar_unit_types = []
        self.columnar_unit_type_classes = []
        self.columnar_unit_id_to_string_mapping = []  # keep this for convenience
        self.columnar_ip_unit_string = []
        self.columnar_preferred_unit_string = []
        self.columnar_units_are_preferred = []  # boolean check
        for row in range(eq.minimum_data_points_for_generation() + 1):
            this_row = []
            if row == 0:
                for col in range(len(column_units)):
                    this_column_unit_type = column_units[col]
                    this_column_unit_type_class = unit_class_factory(this_column_unit_type)
                    self.columnar_unit_type_classes.append(this_column_unit_type_class)
                    this_column_unit_id_to_string_mapping = this_column_unit_type_class.get_unit_string_map()
                    this_column_unit_strings = list(this_column_unit_id_to_string_mapping.values())
                    this_column_preferred_unit_id = this_column_unit_type_class.calculation_unit_id()
                    preferred_unit_string = this_column_unit_id_to_string_mapping[this_column_preferred_unit_id]
                    self.columnar_unit_types.append(this_column_unit_type)
                    self.columnar_unit_id_to_string_mapping.append(this_column_unit_id_to_string_mapping)
                    self.columnar_preferred_unit_string.append(preferred_unit_string)
                    self.columnar_ip_unit_string.append(
                        this_column_unit_id_to_string_mapping[this_column_unit_type_class.base_ip_unit_id()]
                    )
                    self.columnar_units_are_preferred.append(True)
                    this_row.append(f"{preferred_unit_string}")
                    self.table.create_dropdown(
                        r=row,
                        c=col,
                        values=this_column_unit_strings,
                        set_value=preferred_unit_string,
                        state="readonly",
                        redraw=False,
                        selection_function=self._units_changed,
                        modified_function=None
                    )
            else:
                for col in range(len(column_units)):
                    this_row.append(0.0)
            pretend_data.append(this_row)
        self._num_bad_cells = 0
        self.table.headers(column_titles)
        self.table.set_sheet_data(pretend_data)
        self.table.enable_bindings()
        self.table.set_options(expand_sheet_if_paste_too_big=True, paste_insert_column_limit=0)
        self.table.hide(canvas="row_index")
        self.table.hide(canvas="top_left")
        self.table.set_all_cell_sizes_to_text(redraw=True)
        self.table.extra_bindings("end_edit_cell", func=self._cell_edited)
        self.table.extra_bindings("end_paste", func=self._cells_pasted)
        # Undo events in the table could bypass our bindings, need to verify this
        self.table.select_cell(row=1, column=0)
        self.table.pack(side=TOP, expand=True, fill=BOTH, padx=p, pady=p)
        self.table.focus()
        self.expected_num_columns = self.table.total_columns()  # save this to verify later
        tabular_frame.pack(side=TOP, fill=BOTH, expand=True, padx=p, pady=p)
        Label(
            self,
            text="""Once the numeric data is in place, set the numeric units for each column, conform the data if
         necessary, and continue the wizard."""
        ).pack(
            side=TOP, fill=X, expand=False, padx=p, pady=p
        )

        #
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=p, pady=p)
        #
        bottom_button_frame = Frame(self)
        table_ops_label_frame = LabelFrame(bottom_button_frame, text="Table Operations")
        Button(
            table_ops_label_frame, text="Add Table Rows", command=self._add_more_table_rows
        ).grid(
            row=0, column=0, padx=p, pady=p
        )
        Button(
            table_ops_label_frame, text="Shrink Table to Data", command=self._shrink_table_to_data
        ).grid(
            row=0, column=1, padx=p, pady=p
        )
        table_ops_label_frame.grid(row=0, column=0, padx=p, pady=p)
        # Button(
        #     bottom_button_frame, text="Repair Data Column Order", state=DISABLED, command=self._repair
        # ).grid(
        #     row=0, column=1, padx=p, pady=p
        # )
        quick_convert_label_frame = LabelFrame(bottom_button_frame, text="Quick Convert Units")
        Button(
            quick_convert_label_frame, text="IP Units", command=self._quick_convert_ip
        ).grid(
            row=0, column=0, padx=p, pady=p
        )
        Button(
            quick_convert_label_frame, text="SI Units", command=self._quick_convert_si
        ).grid(
            row=0, column=1, padx=p, pady=p
        )
        quick_convert_label_frame.grid(row=0, column=1, padx=p, pady=p)
        workflow_control_label_frame = LabelFrame(bottom_button_frame, text="Primary Controls")
        self.done_conform_text = StringVar(value="Done, Continue")
        Button(
            workflow_control_label_frame, textvariable=self.done_conform_text, command=self._done_or_conform
        ).grid(
            row=0, column=3, padx=p, pady=p
        )
        Button(
            workflow_control_label_frame, text="Cancel Wizard", command=self.cancel
        ).grid(row=0, column=4, padx=p, pady=p)
        workflow_control_label_frame.grid(row=0, column=2, padx=p, pady=p)
        bottom_button_frame.grid_columnconfigure(ALL, weight=1)
        bottom_button_frame.pack(side=TOP, fill=X, expand=False, padx=p, pady=p)

        # set up keybinding and flags
        self.bind('<Key>', self._handle_button_pressed)

        # finalize UI operations
        self.grab_set()
        self.transient(parent_window)

    def _handle_button_pressed(self, event):
        # relevant_modifiers
        # mod_shift = 0x1
        mod_control = 0x4
        # mod_alt = 0x20000
        if event.keysym == 'a' and mod_control & event.state:
            self._add_more_table_rows()
        elif event.keysym == 'd' and mod_control & event.state:
            for row in range(self.table.total_rows()):
                if row == 0:
                    continue
                for col in range(self.table.total_columns()):
                    self.table.set_cell_data(row, col, (row + 1.0) * (col + 1.0))
            self.table.redraw()
            self._cells_pasted()

    def _cell_edited(self, event):
        row = event.row
        if row == 0:
            return
        column = event.column
        # first check the current value to see if it was already bad -- it will be the old value in the cell
        current_value = self.table.get_cell_data(row, column)
        try:
            float(current_value)
            was_good = True
        except ValueError:
            was_good = False
        try:
            float(event.text)
            self.table.highlight_cells(row, column, bg='white')
            # if it was good already, and good now, we are done
            # if it was bad, and now good, decrement the counter
            if not was_good:
                self._num_bad_cells -= 1
        except ValueError:
            self.table.highlight_cells(row, column, bg='pink')
            # if it was bad already, and bad now, we are done
            # if it was good, and now bad, increment the counter
            if was_good:
                self._num_bad_cells += 1
        self.table.redraw()

    def _cells_pasted(self, _=None):
        self._num_bad_cells = 0
        for row in range(self.table.total_rows()):
            if row == 0:
                continue
            for column in range(self.table.total_columns()):
                try:
                    float(self.table.get_cell_data(row, column))
                    self.table.highlight_cells(row, column, bg='white')
                except ValueError:
                    self._num_bad_cells += 1
                    self.table.highlight_cells(row, column, bg='pink')
        self.table.redraw()

    def _table_data_all_good(self) -> bool:
        return self._num_bad_cells == 0

    def _add_more_table_rows(self):
        self.table.insert_rows(rows=10, redraw=True)

    def _shrink_table_to_data(self):
        # first find the extents of the data
        current_visible_rows = self.table.total_rows()
        max_row_index_found = 0
        for row in range(current_visible_rows):
            for col in range(self.table.total_columns()):
                if self.table.get_cell_data(row, col) != '':
                    max_row_index_found = max(max_row_index_found, row)
        # constrain it to the minimums
        minimum_rows = 2  # with the use of correction factors, we are ok allowing just one row here
        desired_row_count = max(max_row_index_found, minimum_rows)
        # if table is larger than desired, start at the end and remove until we get to the desired value
        if current_visible_rows > desired_row_count:
            for r in range(current_visible_rows - 1, desired_row_count, -1):
                self.table.delete_row(r)
        self.table.redraw()

    def any_blank_cells(self):
        for row in range(self.table.total_rows()):
            for col in range(self.table.total_columns()):
                if self.table.get_cell_data(row, col) == '':
                    return True
        return False

    # def _repair(self): pass

    def _quick_convert_ip(self):
        for col_num in range(self.table.total_columns()):
            self.table.set_cell_data(0, col_num, self.columnar_ip_unit_string[col_num], redraw=True)
        self._units_changed()

    def _quick_convert_si(self):
        for col_num in range(self.table.total_columns()):
            self.table.set_cell_data(0, col_num, self.columnar_preferred_unit_string[col_num], redraw=True)
        self._units_changed()

    def _units_changed(self, dropdown_edit_cell_event=None):
        this_column = dropdown_edit_cell_event.column if dropdown_edit_cell_event else None
        self.need_to_conform_units = False
        for c in range(self.table.total_columns()):
            if c == this_column:
                units_value_this_column = dropdown_edit_cell_event.text  # currently being modified
            else:
                units_value_this_column = self.table.get_cell_data(0, c)  # just get the data from the cell
            if units_value_this_column != self.columnar_preferred_unit_string[c]:
                self.need_to_conform_units = True
        self._cells_pasted()  # simulate a paste event
        self.refresh_done_conform_button_text()

    def refresh_done_conform_button_text(self):
        if self.need_to_conform_units:
            self.done_conform_text.set("Conform units")
        else:
            self.done_conform_text.set("Done, Continue")
        self.table.redraw()

    def _update_from_traces(self):
        pass

    def _done_or_conform(self):
        # don't try to conorm or exit if there are any bad values
        if not self._table_data_all_good():
            message_window = PetMessageForm(
                self, "Value Issue", f"{self._num_bad_cells} cell(s) appear to have bad values; fix them and retry"
            )
            self.wait_window(message_window)
            return
        if self.need_to_conform_units:
            self.conform_units()
        else:
            if self.expected_num_columns != self.table.total_columns():
                message_window = PetMessageForm(
                    self,
                    "Problem with Table Size",
                    "It appears the table column size does not match the expected size, this is odd, cancel and retry"
                )
                self.wait_window(message_window)
                return
            if self.any_blank_cells():
                message_window = PetMessageForm(
                    self,
                    "Problem with Table Entries",
                    "Blank entries detected in the table, fix and retry"
                )
                self.wait_window(message_window)
                return
            self.exit_code = MainDataForm.MainDataExitCode.Done
            self.final_base_data_rows = []
            for row in range(self.table.total_rows()):
                if row == 0:  # unit row
                    continue
                this_row = []
                for column in range(self.table.total_columns()):
                    this_row.append(float(self.table.get_cell_data(row, column)))
                self.final_base_data_rows.append(this_row)
            self.grab_release()
            self.destroy()

    def conform_units(self):
        """This is a pretty convoluted way to do this, think harder"""
        for c in range(self.table.total_columns()):
            current_units_string = self.table.get_cell_data(0, c)
            if current_units_string != self.columnar_preferred_unit_string[c]:
                try:
                    current_unit_id = self.columnar_unit_type_classes[c].get_id_from_unit_string(current_units_string)
                except EnergyPlusPetException:
                    pmf = PetMessageForm(self, "Unit ID Problem", "Could not match unit ID; this is a developer issue.")
                    self.wait_window(pmf)
                    self.cancel()
                    return
                for r in range(self.table.total_rows()):
                    if r == 0:
                        self.table.set_cell_data(r, c, self.columnar_preferred_unit_string[c])
                    else:
                        cell_value = float(self.table.get_cell_data(r, c))
                        unit_value = unit_instance_factory(cell_value, self.columnar_unit_types[c])
                        unit_value.units = current_unit_id
                        unit_value.convert_to_calculation_unit()
                        self.table.set_cell_data(r, c, unit_value.value)
        self.need_to_conform_units = False
        self.refresh_done_conform_button_text()

    def cancel(self):
        self.exit_code = MainDataForm.MainDataExitCode.Cancel
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_heating_curve import WaterToAirHeatPumpHeatingCurveFit

    root = Tk()
    MainDataForm(root, WaterToAirHeatPumpHeatingCurveFit())
    root.mainloop()



from tkinter import Button, Frame, Label, LabelFrame, TOP, Spinbox, IntVar, Scrollbar, LEFT, BOTH, RIGHT, EW, \
    VERTICAL, Radiobutton, StringVar, W, NS, OptionMenu, MULTIPLE, Listbox, Variable, BooleanVar, \
    ACTIVE, DISABLED, HORIZONTAL, TclError
from tkinter.ttk import Separator
from typing import Callable

from energyplus_pet.correction_factor import CorrectionFactor, CorrectionFactorType
from energyplus_pet.equipment.base import BaseEquipment


class CorrectionSummaryWidget(LabelFrame):
    """
    This widget returns a Tk.LabelFrame that displays a correction factor summary and is shown on the summary form.
    """

    def __init__(self, parent: Frame, name: str, equipment_instance: BaseEquipment, remove_callback: Callable):
        super().__init__(parent, name=name, text=name)
        self.cf = CorrectionFactor(name)
        self.equip_instance = equipment_instance

        # these are Tk variables for tracking dynamic changes and tracing
        self.var_base_column = StringVar(value=self.equip_instance.headers().name_array()[0])
        self.var_num_corrections = IntVar(value=self.cf.num_corrections)
        self.var_wb_db = BooleanVar(value=False)
        self.var_mod_type = StringVar(value=self.cf.correction_type.name)

        # finally build out the gui ahead of setting up the traces
        self._build_gui()

        # set up tracing
        self.var_num_corrections.trace('w', self._update_from_traces)
        self.var_mod_type.trace('w', self._update_from_traces)
        self.var_base_column.trace('w', self._update_from_traces)
        self.var_wb_db.trace('w', self._update_from_traces)

        # other misc GUI functions
        self._setup_removal_callback(remove_callback)

    def _update_from_traces(self, *_):  # pragma: no cover
        try:
            self.cf.num_corrections = self.var_num_corrections.get()
        except TclError:
            self.cf.num_corrections = 0  # if blank, set to zero for now
        if self.var_mod_type.get() == CorrectionFactorType.Multiplier.name:
            self.cf.correction_type = CorrectionFactorType.Multiplier
        elif self.var_mod_type.get() == CorrectionFactorType.Replacement.name:
            self.cf.correction_type = CorrectionFactorType.Replacement
        elif self.var_mod_type.get() == CorrectionFactorType.CombinedDbWb.name:
            self.cf.correction_type = CorrectionFactorType.CombinedDbWb
        mod_column = self.var_base_column.get()
        self.cf.base_column_index = self.equip_instance.headers().name_array().index(mod_column)
        self.cf.columns_to_modify = self.columns_listbox.curselection()
        if self.cf.correction_type == CorrectionFactorType.CombinedDbWb:
            self.base_column_dropdown['state'] = DISABLED
        else:
            self.base_column_dropdown['state'] = ACTIVE

    def _setup_removal_callback(self, remove_callback: Callable):
        self.remove_me = False
        self.remove_callback = remove_callback

    def _build_gui(self):  # pragma: no cover
        p = 4
        Button(self, text="❌ Remove This Factor", command=self._remove).grid(
            row=0, column=0, padx=p, pady=p
        )
        Separator(self, orient=HORIZONTAL).grid(
            row=1, column=0, sticky=EW, padx=p, pady=p
        )
        corr_frame = Frame(self)
        Label(corr_frame, text="# Correction Values").grid(
            row=0, column=0, padx=p, pady=p
        )
        Spinbox(corr_frame, from_=1, to=99, width=4, textvariable=self.var_num_corrections).grid(
            row=0, column=1, padx=p, pady=p
        )
        corr_frame.grid(
            row=2, column=0, padx=p, pady=p
        )
        lf = LabelFrame(self, text="Correction Factor Type")
        Radiobutton(
            lf, text="Multiplier", value=CorrectionFactorType.Multiplier.name, variable=self.var_mod_type
        ).pack(
            side=TOP, anchor=W, padx=p, pady=p
        )
        Radiobutton(
            lf, text="Replacement", value=CorrectionFactorType.Replacement.name, variable=self.var_mod_type
        ).pack(
            side=TOP, anchor=W, padx=p, pady=p
        )
        headers = self.equip_instance.headers()
        state = ACTIVE if headers.get_db_column() >= 0 and headers.get_wb_column() >= 0 else DISABLED
        Radiobutton(
            lf, text="Combined DB/WB Factor", value=CorrectionFactorType.CombinedDbWb.name, variable=self.var_mod_type,
            state=state
        ).pack(
            side=TOP, anchor=W, padx=p, pady=p
        )
        lf.grid(row=3, column=0, columnspan=2, padx=p, pady=p)
        Separator(self, orient=VERTICAL).grid(
            row=0, column=1, rowspan=5, sticky=NS, padx=p, pady=p
        )
        options = Variable(value=self.equip_instance.headers().name_array())
        Label(self, text="Base data column for this correction factor:").grid(
            row=0, column=2, padx=p, pady=p
        )
        self.base_column_dropdown = OptionMenu(self, self.var_base_column, *options.get())
        self.base_column_dropdown.grid(row=1, column=2, sticky=EW, padx=p, pady=p)
        Label(self, text="Data affected by this correction factor:").grid(
            row=2, column=2, padx=p, pady=p
        )
        columns_frame = Frame(self)
        self.columns_listbox = Listbox(columns_frame, height=5, listvariable=options, selectmode=MULTIPLE)
        self.columns_listbox.configure(exportselection=False)
        self.columns_listbox.pack(side=LEFT, fill=BOTH, expand=True, padx=p, pady=3)
        self.columns_listbox.bind('<<ListboxSelect>>', self._update_from_traces)
        columns_scroll = Scrollbar(columns_frame)
        columns_scroll.pack(side=RIGHT, fill=BOTH)
        self.columns_listbox.config(yscrollcommand=columns_scroll.set)
        columns_scroll.config(command=self.columns_listbox.yview)
        columns_frame.grid(
            row=3, column=2, sticky=EW, padx=p, pady=p
        )

    def _remove(self):
        self.remove_me = True
        self.remove_callback()

    def check_ok(self) -> bool:
        """Checks the underlying correction factor for validity"""
        db = self.equip_instance.headers().get_db_column()
        wb = self.equip_instance.headers().get_wb_column()
        return self.cf.check_ok(db, wb, summary_only=True)

    def description(self):
        return f"CorrectionFactor {self.equip_instance.name}; {self.cf.num_corrections} corrections"



from platform import system
from json import dumps
from pathlib import Path
from queue import Queue
from subprocess import check_call
from sys import executable
from threading import Thread
from tkinter import BOTH, LEFT, RIGHT, TOP, BOTTOM, X, Y  # widget sides and directions to use in widget.pack commands
from tkinter import END  # key used when adding data to the scrolledText object
from tkinter import IntVar, StringVar  # GUI variables
from tkinter import NSEW, EW, S  # sticky cardinal directions to use in widget grid commands
from tkinter import SUNKEN, DISABLED, ACTIVE  # attributes used to modify widget appearance
from tkinter import Tk, Button, Frame, Label, PhotoImage, scrolledtext, Scrollbar, Menu  # widgets
from tkinter import messagebox, filedialog  # simple dialogs for user messages
from tkinter.ttk import LabelFrame, Progressbar, Treeview, Separator, Notebook  # ttk widgets
from webbrowser import open as browser_open

from pyshortcuts import make_shortcut

from energyplus_pet import NICE_NAME, VERSION
from energyplus_pet.forms.correction_detail_form import DetailedCorrectionFactorForm
from energyplus_pet.data_manager import CatalogDataManager
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.equipment.manager import EquipmentFactory
from energyplus_pet.equipment.equip_types import EquipType, EquipTypeUniqueStrings as ETString
from energyplus_pet.forms.constant_parameters import ConstantParameterEntryForm
from energyplus_pet.forms.correction_summary_form import CorrectionFactorSummaryForm
from energyplus_pet.forms.header_preview import RequiredDataPreviewForm
from energyplus_pet.forms.catalog_plot import CatalogDataPlotForm
from energyplus_pet.forms.comparison_plot import ComparisonPlot
from energyplus_pet.forms.base_data_form import MainDataForm


class EnergyPlusPetWindow(Tk):
    """This form is the primary GUI entry point for the program; all control runs through here"""

    def __init__(self):
        """
        The main window of the parameter estimation tool GUI workflow.
        This window is an instance of a tk.Tk object
        """
        super().__init__()

        # set some basic program information like title and an icon
        self._program_name = NICE_NAME
        program_name_with_version = f"{self._program_name} {VERSION}"
        self.title(program_name_with_version)
        icon_path = Path(__file__).parent / 'favicon.png'
        image = PhotoImage(file=str(icon_path))
        self.iconphoto(True, image)

        # setup event listeners
        self._gui_queue = Queue()
        self._check_queue()

        # define the Tk.Variable instances that will be used to communicate with the GUI widgets
        self._define_tk_variables()

        # build out the form and specify a minimum size, which may not be uniform across platforms
        self._build_gui()
        self.update()
        self.minsize(self.winfo_width(), self.winfo_height())

        # set up some important member variables
        self._equip_instance: BaseEquipment = BaseEquipment()  # sits as a dummy value initially, may not be necessary
        self._catalog_data_manager = CatalogDataManager()
        self._thread_running = False

        # window setup operations
        self._update_status_bar("Program Initialized")
        self._refresh_gui_state()
        self.bind('<Key>', self._handle_button_pressed)

    def _handle_button_pressed(self, event):
        # relevant_modifiers
        # mod_shift = 0x1
        mod_control = 0x4
        # mod_alt = 0x20000
        if event.keysym == 'e' and mod_control & event.state:
            if self._button_engage['state'] != DISABLED:
                self._engage()
        elif event.keysym == 'r' and mod_control & event.state:
            if self._button_preview['state'] != DISABLED:
                self._preview_data()
        elif event.keysym == 't' and mod_control & event.state:
            if self._button_catalog['state'] != DISABLED:
                self._catalog_data_wizard()
        elif event.keysym == 's' and mod_control & event.state:
            if self._button_save_data['state'] != DISABLED:
                self._save_data_to_file()

    def _check_queue(self):
        """Checks the GUI queue for actions and sets a timer to check again each time"""
        while True:
            # noinspection PyBroadException
            try:
                task = self._gui_queue.get(block=False)
                self.after_idle(task)
            except Exception:
                break
        self.after(100, self._check_queue)

    def _define_tk_variables(self):
        """Creates and initializes all the Tk.Variable instances used in the GUI for two-way communication"""
        self._tk_var_progress = IntVar(value=0)
        self._tk_var_status_equip = StringVar(value="Selected Equipment: NONE")
        self._tk_var_status_data = StringVar(value="Catalog Data: NOT READY")
        self._tk_var_status_status = StringVar(value="Program Initialized")

    def _build_gui(self):
        """Builds out the entire window GUI, calling workers as necessary"""
        # now build the top menubar, it's not part of the geometry, but a config parameter on the root Tk object
        self._build_menu()

        # create an instance of the equipment id/tree mapping and then build the tree GUI contents
        label_frame_equip_type = LabelFrame(self, text="Equipment Type")
        self._build_treeview(label_frame_equip_type)
        label_frame_equip_type.grid(row=0, column=0, sticky=NSEW)

        # build the options/controls GUI contents
        label_frame_control = LabelFrame(self, text="Options and Controls")
        self._build_controls(label_frame_control)
        label_frame_control.grid(row=0, column=1, sticky=NSEW)

        # build the output GUI contents
        label_frame_output = LabelFrame(self, text="Parameter Output")
        self._build_output(label_frame_output)
        label_frame_output.grid(row=0, column=2, sticky=NSEW)

        # build the status bar
        status_frame = Frame(self)
        Label(status_frame, relief=SUNKEN, anchor=S, textvariable=self._tk_var_status_equip).pack(
            side=LEFT, fill=BOTH, expand=True
        )
        Label(status_frame, relief=SUNKEN, anchor=S, textvariable=self._tk_var_status_data).pack(
            side=LEFT, fill=BOTH, expand=True
        )
        Label(status_frame, relief=SUNKEN, anchor=S, textvariable=self._tk_var_status_status).pack(
            side=LEFT, fill=BOTH, expand=True
        )
        status_frame.grid(row=1, column=0, columnspan=3, sticky=EW)

        # set up the weight of each row (even) and column (distributed) for a nice looking GUI
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        self.grid_columnconfigure(2, weight=8)

    def _build_menu(self):
        """Builds out the menubar at the top of thw window"""
        menubar = Menu(self)
        menu_help = Menu(menubar, tearoff=0)
        menu_help.add_command(label="Open online documentation...", command=self._help_documentation)
        menu_help.add_command(label="Open examples folder...", command=self._open_examples)
        menu_help.add_command(label="Create desktop icon", command=self._create_shortcut)
        menu_help.add_command(label="About...", command=self._help_about)
        menubar.add_cascade(label="Help", menu=menu_help)
        self.config(menu=menubar)

    def _build_treeview(self, container):
        """Builds out the equipment treeview on the left side of the window"""
        tree_holder = Frame(container)
        equip_type_scrollbar = Scrollbar(tree_holder)
        # I don't believe the Treeview object can interact with a Tk.Variable, so the tree is stored on the class
        self._tree = Treeview(tree_holder, columns=('Type',), show='tree', yscrollcommand=equip_type_scrollbar.set)
        self._tree.column('#0', stretch=False)
        equip_type_scrollbar.config(command=self._tree.yview)
        # eventually use a defined dictionary somewhere for the tree items and keywords
        root_hp = self._tree.insert(parent='', index='end', text="Heat Pump Coils", open=True)
        branch_wah = self._tree.insert(parent=root_hp, index='end', text="Water to Air Heating Coil", open=True)
        self.init = self._tree.insert(
            parent=branch_wah, index='end', text="Curve Fit", tags=ETString.WAHP_Heating_CurveFit
        )
        # self._tree.insert(parent=branch_wah, index='end', text="Parameter Estimation", tags=ETString.WAHP_Heating_PE)
        branch_wac = self._tree.insert(parent=root_hp, index='end', text="Water to Air Cooling Coil", open=True)
        self._tree.insert(parent=branch_wac, index='end', text="Curve Fit", tags=ETString.WAHP_Cooling_CurveFit)
        # self._tree.insert(parent=branch_wac, index='end', text="Parameter Estimation", tags=ETString.WAHP_Cooling_PE)
        branch_wwh = self._tree.insert(parent=root_hp, index='end', text="Water to Water Heating", open=True)
        self._tree.insert(parent=branch_wwh, index='end', text="Curve Fit", tags=ETString.WWHP_Heating_CurveFit)
        branch_wwc = self._tree.insert(parent=root_hp, index='end', text="Water to Water Cooling", open=True)
        self._tree.insert(parent=branch_wwc, index='end', text="Curve Fit", tags=ETString.WWHP_Cooling_CurveFit)
        # root_pumps = self._tree.insert(parent='', index='end', text='Pumps', open=True)
        # branch_con_pump = self._tree.insert(parent=root_pumps, index='end', text="Constant Speed Pump", open=True)
        # self._tree.insert(
        #     parent=branch_con_pump, index='end', text='Non-Dimensional', tags=ETString.Pump_ConstSpeed_ND
        # )
        self._tree.pack(side=LEFT, padx=3, pady=3, fill=BOTH, expand=True)
        equip_type_scrollbar.pack(side=RIGHT, padx=0, pady=3, fill=Y, expand=False)
        self._tree.focus(self.init)
        self._tree.selection_set(self.init)
        tree_holder.pack(side=TOP, fill=BOTH, expand=True)

    def _build_controls(self, container):
        """Builds out the control section in the middle of the window"""
        Label(
            container, text="Select an equipment type in the tree\nand press Ctrl-e or click here:"
        ).pack(side=TOP, padx=3, pady=3)
        self._button_engage = Button(container, text='Engage Equipment Type', command=self._engage)
        self._button_engage.pack(side=TOP, padx=3, pady=3, fill=X)
        Separator(container, orient='horizontal').pack(fill=X, padx=3, pady=3)
        Label(
            container, text="To view the required data for the selected\nequipment, press Ctrl-r or click here:"
        ).pack(side=TOP, padx=3, pady=3)
        self._button_preview = Button(container, text="Required Data Description", command=self._preview_data)
        self._button_preview.pack(side=TOP, padx=3, pady=3, fill=X)
        Separator(container, orient='horizontal').pack(fill=X, padx=3, pady=3)
        Label(
            container, text="Finally, to enter data and process parameters\n press Ctrl-t or click here..."
        ).pack(side=TOP, padx=3, pady=3)
        self._button_catalog = Button(container, text="Catalog Data Wizard", command=self._catalog_data_wizard)
        self._button_catalog.pack(side=TOP, padx=3, pady=3, fill=X)
        Label(container, text="...and watch the progress here:").pack(side=TOP, padx=3, pady=3)
        self._progress = Progressbar(container, variable=self._tk_var_progress)
        self._progress.pack(side=TOP, padx=3, pady=3, fill=X)
        Separator(container, orient='horizontal').pack(fill=X, padx=3, pady=3)
        Label(container, text="Once complete, use Ctrl-s to save data:").pack(side=TOP, padx=3, pady=3)
        self._button_save_data = Button(
            container, text="Save Output to File", command=self._save_data_to_file, state=DISABLED,
        )
        self._button_save_data.pack(side=TOP, fill=X, padx=3, pady=3)
        Separator(container, orient='horizontal').pack(fill=X, padx=3, pady=3)
        Label(container, text="Then if you want to run another, just reinitialize here:").pack(
            side=TOP, padx=3, pady=3)
        Button(container, text="Reinitialize Form", command=self._reinitialize).pack(side=TOP, fill=X, padx=3, pady=3)

    @staticmethod
    def _build_one_output_frame(notebook_container: Notebook, tab_title: str):
        """Builds out a single output frame to be in the output notebook"""
        output_frame_par = Frame(notebook_container)
        output_frame_par.pack(side=TOP, expand=True, fill=BOTH)
        horizontal_scroller_par = Scrollbar(output_frame_par, orient='horizontal')
        horizontal_scroller_par.pack(side=BOTTOM, fill=X)
        this_output_box = scrolledtext.ScrolledText(
            output_frame_par, wrap='none', width=40, height=20, xscrollcommand=horizontal_scroller_par.set
        )
        this_output_box.pack(side=TOP, padx=3, pady=3, fill=BOTH, expand=True)
        this_output_box.insert(END, 'Messages and results will appear here')
        horizontal_scroller_par.config(command=this_output_box.xview)
        notebook_container.add(output_frame_par, text=tab_title)
        return this_output_box

    def _build_output(self, container):
        """Builds out the full output notebook on the right side of the window, calling workers as needed"""
        output_notebook = Notebook(container)
        self._output_box_par = self._build_one_output_frame(output_notebook, "Parameter Summary")
        self._output_box_idf = self._build_one_output_frame(output_notebook, "IDF Object")
        self._output_box_json = self._build_one_output_frame(output_notebook, "EpJSON Object")
        output_notebook.pack(expand=True, fill=BOTH)

    def _refresh_gui_state(self):
        """Checks current instance flags and sets button states appropriately"""
        if self._thread_running:
            self._button_engage['state'] = DISABLED
            self._button_catalog['state'] = DISABLED
            self._button_preview['state'] = DISABLED
            self._button_save_data['state'] = ACTIVE
        elif self._equip_instance is None or self._equip_instance.this_type() == EquipType.InvalidType:
            self._button_engage['state'] = ACTIVE
            self._button_catalog['state'] = DISABLED
            self._button_preview['state'] = DISABLED
            self._button_save_data['state'] = DISABLED
        else:
            self._button_engage['state'] = ACTIVE
            self._button_catalog['state'] = ACTIVE
            self._button_preview['state'] = ACTIVE
            if self._catalog_data_manager.data_processed:
                self._button_save_data['state'] = ACTIVE
            else:
                self._button_save_data['state'] = DISABLED

    def _help_documentation(self):
        """Launches a browser to open the stable documentation"""
        # could try to use the current version docs but that may be a bit finicky
        browser_open('https://energypluspet.readthedocs.io/en/stable/')
        self._update_status_bar('Launched online documentation')

    def _create_shortcut(self):
        runner_script = str(Path(__file__).resolve().parent.parent / 'runner.py')
        icon_extension = 'ico' if system() == 'Windows' else 'png'
        icon_path = str(Path(__file__).resolve().parent / f"favicon.{icon_extension}")
        make_shortcut(
            runner_script, name=self._program_name, terminal=False, icon=icon_path,
            executable=executable, startmenu=False
        )

    @staticmethod
    def _open_examples():
        examples_dir = str(Path(__file__).resolve().parent.parent / 'examples')
        if system() == 'Darwin':
            check_call(["open", examples_dir])
        elif system() == 'Linux':
            check_call(["xdg-open", examples_dir])
        elif system() == 'Windows':
            from os import startfile
            startfile(examples_dir)

    def _preview_data(self):
        """Allows the user to preview the data for the selected equipment, button is disabled until an equip is set."""
        preview = RequiredDataPreviewForm(self, self._equip_instance)
        self.wait_window(preview)

    def _help_about(self):
        """Simple about window showing program information"""
        messagebox.showinfo(
            f"About {self._program_name}",
            "The original idea for this tool was developed in 2009ish by Edwin Lee and implemented in a VB.Net tool.\n"
            "This new version is developed in Python to allow for easy cross platform development and packaging."
        )

    def _save_data_to_file(self):
        """Saves catalog data to a file"""
        file_path = filedialog.asksaveasfilename(
            parent=self, filetypes=(('JSON File', '.json'),), confirmoverwrite=True
        )
        if file_path is None or file_path == ():
            return
        try:
            with open(file_path, 'w') as f:
                response_object = {
                    'catalog_inputs': self._catalog_data_manager.summary(),
                    'parameter_summary': self._output_box_par.get('1.0', END),
                    'idf_object': self._output_box_idf.get('1.0', END),
                    'epjson_object': self._output_box_json.get('1.0', END),
                }
                f.write(dumps(response_object, indent=2))
        except Exception as e:  # noqa  any file issue could happen
            messagebox.showerror(self._program_name, "Could not save data to file")

    def _reinitialize(self):
        self._update_all_output_boxes('Messages and results will appear here')
        self._equip_instance = None
        self._catalog_data_manager.reset()
        self._tk_var_progress.set(0)
        self._tree.focus(self.init)
        self._tree.selection_set(self.init)
        self._refresh_gui_state()

    def _engage(self) -> None:
        """
        Checks the currently selected object in the equipment tree, and if valid, assigns the
        self.selected_equip_instance member variable to a proper equipment instance.
        GUI state is refreshed if anything changed

        :return: Nothing
        """
        cur_item = self._tree.focus()
        node_tags = self._tree.item(cur_item, 'tags')  # this always either returns '' or ('tag_contents',)
        if node_tags == '':
            messagebox.showwarning("Type Issue", "Make sure to select one of the child types, not a parent tree node")
            return
        node_tag = node_tags[0]
        potential_new_equip_type = ETString.get_equip_type_from_unique_string(node_tag)
        if not self._catalog_data_manager.data_processed:
            # then we are just selecting a new equip type, select it and move on
            potential_equip = EquipmentFactory.instance_factory(potential_new_equip_type)
            if potential_equip is None:
                messagebox.showwarning("Not Implemented Yet", "This type has not been implemented yet, sorry!")
                return
            self._equip_instance = potential_equip
            self._catalog_data_manager.reset()
            self._tk_var_progress.set(0)
            self._update_status_bar("New Equipment Type Selected")
            self._refresh_gui_state()
            return
        if potential_new_equip_type == self._equip_instance.this_type():
            messagebox.showinfo("Type Issue", "This equipment type was already engaged, not making any changes")
            return
        else:
            response = messagebox.askyesno(
                "Type Issue",
                "New type selected, but catalog data is already present. Would you like to clear the old data?"
            )
            if response:
                self._catalog_data_manager.reset()
                self._update_status_bar("New Equipment Type Selected")
            self._refresh_gui_state()

    def _catalog_data_wizard(self):
        """
        Manages the data entry processes.  Calls child functions to display forms, handles different exits, etc.
        This function assumes the self.selected_equip_instance has already been set to a proper instance in the engage
        function.

        :return: Nothing
        """
        self._update_status_bar('Started catalog wizard')
        self._update_all_output_boxes('Catalog wizard starting')

        self._tk_var_progress.set(0)
        self._progress['maximum'] = self._equip_instance.get_number_of_progress_steps() + 9  # 9 steps in here
        self._handler_thread_increment()

        # first open a correction factor definition form window, if this returns False, it means abort
        cdf = CorrectionFactorSummaryForm(self, self._equip_instance)
        self.wait_window(cdf)
        if cdf.exit_code == CorrectionFactorSummaryForm.ExitCode.Cancel:
            # in the original code, this would set CorrectionsExist to false, not sure that's right
            self._update_status_bar('Correction factor form cancelled')
            self._update_all_output_boxes('Correction factor form cancelled; reopen wizard to try again')
            self._tk_var_progress.set(0)
            return  # correction data form was cancelled, just abort
        elif cdf.exit_code == CorrectionFactorSummaryForm.ExitCode.Error:
            self._update_status_bar('Correction factor form error!')
            self._update_all_output_boxes('Correction factor form encountered an error; reopen wizard to try again')
            self._tk_var_progress.set(0)
            return  # may need to report the error
        else:
            self._update_status_bar('Correction factor summary complete')
        self._handler_thread_increment()

        # if that was successful, loop over each local summary and open individual correction entry forms for each
        for cf_num, cf in enumerate(cdf.factor_summaries):
            cfd_form = DetailedCorrectionFactorForm(
                self, cf, self._equip_instance, cf_num + 1, len(cdf.factor_summaries)
            )
            self.wait_window(cfd_form)
            if cfd_form.exit_code == DetailedCorrectionFactorForm.DetailedCorrectionExitCode.Cancel:
                self._update_status_bar('Correction factor form cancelled')
                self._update_all_output_boxes('Correction factor form cancelled; reopen wizard to try again')
                self._tk_var_progress.set(0)
                return
            self._catalog_data_manager.add_correction_factor(cfd_form.completed_factor)
        self._update_status_bar('Correction factor details complete')
        self._handler_thread_increment()

        # now that we have the full correction factor details, we need to collect the main catalog data
        main_catalog_data_form = MainDataForm(self, self._equip_instance)
        self.wait_window(main_catalog_data_form)
        if main_catalog_data_form.exit_code == MainDataForm.MainDataExitCode.Cancel:
            self._update_status_bar('Main catalog data form cancelled')
            self._update_all_output_boxes('Main catalog data form cancelled; reopen wizard to try again')
            self._tk_var_progress.set(0)
            return
        self._catalog_data_manager.add_base_data(main_catalog_data_form.final_base_data_rows)
        self._handler_thread_increment()

        # then process the base data and correction factors into a full data set
        response_status = self._catalog_data_manager.apply_correction_factors(
            self._equip_instance.minimum_data_points_for_generation(),
            self._equip_instance.headers().get_db_column(),
            self._equip_instance.headers().get_wb_column()
        )
        if response_status == CatalogDataManager.ProcessResult.ERROR:
            self._update_status_bar('Error processing catalog data')
            self._update_all_output_boxes(
                f"Error processing data! Message: \n\n{self._catalog_data_manager.last_error_message}"
            )
            self._tk_var_progress.set(0)
            return
        self._handler_thread_increment()

        # if this equipment requires constant/rated parameters, get them now
        if len(self._equip_instance.get_required_constant_parameters()) > 0:
            cde = ConstantParameterEntryForm(self, self._equip_instance)
            self.wait_window(cde)
            if cde.form_cancelled:
                self._update_status_bar('Constant parameter data form cancelled')
                self._update_all_output_boxes('Constant parameter data form cancelled; reopen wizard to try again')
                self._tk_var_progress.set(0)
                return
            for k, v in cde.parameter_value_map.items():
                self._equip_instance.set_required_constant_parameter(k, v)
        self._handler_thread_increment()

        # and actually, if the data doesn't have diversity, we should accept it, but not allow creating parameters
        # the user should be able to reopen the wizard and add more data to variables or whatever
        # then display the catalog data plot form for inspection
        cdf = CatalogDataPlotForm(self, self._catalog_data_manager, self._equip_instance)
        cdf.wait_window()
        if cdf.exit_code == CatalogDataPlotForm.ExitCode.CANCEL:
            self._update_status_bar('Catalog data plot form cancelled')
            self._update_all_output_boxes('Catalog data plot form cancelled; reopen wizard to try again')
            self._tk_var_progress.set(0)
            return
        self._handler_thread_increment()

        # catalog data is now ready
        self._update_status_bar('Processed Catalog Data')
        self._update_all_output_boxes('Catalog data plot form cancelled; reopen wizard to try again')
        self._handler_thread_increment()

        # the parameter generation process could be lengthy, so we put it in a background thread
        self._update_status_bar('Starting parameter generation process')
        self._thread_running = True
        self._refresh_gui_state()
        thd = Thread(target=self._worker_generate_params, args=(self._equip_instance, self._catalog_data_manager))
        thd.daemon = True
        thd.start()
        self._handler_thread_increment()

    def _worker_generate_params(self, equip_instance: BaseEquipment, data: CatalogDataManager) -> None:
        """
        Function that will be in a background thread, calls the equipment to generate parameters.
        Equipment instance passed in will be mutated so that when the main thread accesses it, it will be updated.

        :param equip_instance: An equipment instance constructed by the main form
        :param data: A fully populated catalog data manager, with final data ready for the equipment to process
        :return: Nothing
        """
        try:
            equip_instance.generate_parameters(data, self._callback_thread_increment, self._callback_thread_done)
        except Exception as e:  # any type of exception
            self._callback_thread_done(False, "Error occurred! " + str(e))
            return

    def _update_status_bar(self, extra_message: str) -> None:
        """
        Updates the status bar at the bottom of the window, providing data based on flags and displaying the message

        :param extra_message: String message to show on the right side of the status bar
        :return: Nothing
        """
        if self._equip_instance is None:
            self._tk_var_status_equip.set("Selected Equipment: NONE")
        else:
            self._tk_var_status_equip.set(f"Selected Equipment: {self._equip_instance.short_name()}")
        if not self._catalog_data_manager.data_processed:
            self._tk_var_status_data.set("Catalog data: NOT READY")
        else:
            self._tk_var_status_data.set("Catalog Data: READY")
        self._tk_var_status_status.set(extra_message)

    def _handler_thread_increment(self) -> None:
        """Main thread handler for a progress increment step"""
        self._tk_var_progress.set(self._tk_var_progress.get() + 1)

    def _callback_thread_increment(self) -> None:
        """Background thread callback function for inserting a progress increment step in the GUI queue"""
        self._gui_queue.put(self._handler_thread_increment)

    def _update_par_box(self, content: str) -> None:
        """
        Minimal worker function to update the data in the parameter summary output box

        :param content: String data to display in the output box
        :return: Nothing
        """
        self._output_box_par.delete('1.0', END)
        self._output_box_par.insert(END, content)

    def _update_idf_box(self, content: str) -> None:
        """
        Minimal worker function to update the data in the IDF Object output box

        :param content: String data to display in the output box
        :return: Nothing
        """
        self._output_box_idf.delete('1.0', END)
        self._output_box_idf.insert(END, content)

    def _update_json_box(self, content: str) -> None:
        """
        Minimal worker function to update the data in the EpJSON output box

        :param content: String data to display in the output box
        :return: Nothing
        """
        self._output_box_json.delete('1.0', END)
        self._output_box_json.insert(END, content)

    def _update_all_output_boxes(self, content: str) -> None:
        """
        Minimal worker function to update the data in all output boxes at once

        :param content: String data to display in the output boxes
        :return: Nothing
        """
        self._update_par_box(content)
        self._update_idf_box(content)
        self._update_json_box(content)

    def _handler_thread_done(self, success: bool, err_message: str = '') -> None:
        """
        Main thread handler for the background thread being completed

        :param success: Boolean flag for whether the process was a success or not
        :param err_message: Optional error message in case something went wrong
        :return: Nothing
        """
        self._thread_running = False
        self._refresh_gui_state()
        self._update_status_bar('Finished Parameter Generation')
        if success:
            self._update_par_box(self._equip_instance.to_parameter_summary())
            self._update_idf_box(self._equip_instance.to_eplus_idf_object())
            self._update_json_box(self._equip_instance.to_eplus_epjson_object())
            self.wait_window(ComparisonPlot(self, self._catalog_data_manager, self._equip_instance))
            self._tk_var_progress.set(100)
        else:
            self._update_all_output_boxes(err_message)

    def _callback_thread_done(self, success: bool, err_message: str = '') -> None:
        """Background thread callback function for inserting a progress done step in the GUI queue"""
        self._gui_queue.put(lambda: self._handler_thread_done(success, err_message))

    def run(self) -> None:
        """Executes the Tk main loop to handle all GUI events and update"""
        self.mainloop()



from typing import List

from energyplus_pet.units import UnitType


class ColumnHeader:
    """Defines a single column of data for an equipment type"""
    def __init__(self, column_name: str, unit_type: UnitType, db: bool = False, wb: bool = False):
        """
        Constructs a single column header

        :param column_name: The string name of this column
        :param unit_type: The UnitType for this column
        :param db: An optional bool flag for whether this column is an air dry-bulb column
        :param wb: An optional bool flag for whether this column is an air wet-bulb column
        """
        self.name = column_name
        self.units_type = unit_type
        self.dry_bulb_column_flag = db
        self.wet_bulb_column_flag = wb


class ColumnHeaderArray:
    """Defines a full set of headers for an equipment type"""
    def __init__(self, columns: List[ColumnHeader]):
        """Constructs an array of columns from a list of column headers"""
        self.columns = columns

    def __len__(self) -> int:
        """Returns the length of the header array, so it can be used in len() expressions"""
        return len(self.columns)

    def unit_array(self) -> List[UnitType]:
        """Returns the list of units for these columns"""
        return [c.units_type for c in self.columns]

    def name_array(self) -> List[str]:
        """Returns the list of strings for these columns"""
        return [c.name for c in self.columns]

    def get_descriptive_summary(self) -> str:
        """Returns a descriptive summary of this set of column headers"""
        response = ""
        for c in self.columns:
            response += f"{c.name} [{c.units_type}]\n"
        return response

    def get_descriptive_csv(self) -> str:
        """Returns a CSV summary of this set of column headers"""
        name_row = ','.join(self.name_array())
        unit_row = ','.join([str(u) for u in self.unit_array()])
        return f"{name_row}\n{unit_row}"

    def get_db_column(self) -> int:
        """Returns a zero-based index of the air dry-bulb column in a set of headers"""
        for i, c in enumerate(self.columns):
            if c.dry_bulb_column_flag:
                return i
        return -1

    def get_wb_column(self) -> int:
        """Returns a zero-based index of the air wet-bulb column in a set of headers"""
        for i, c in enumerate(self.columns):
            if c.wet_bulb_column_flag:
                return i
        return -1



from enum import Enum, auto
import numpy as np
from tkinter import Toplevel, Frame  # containers
from tkinter import Button, Label  # widgets
from tkinter import TOP, X, BOTH, ALL  # appearance stuff
from tkinter.ttk import Notebook, Style  # ttk specific stuff

from energyplus_pet.data_manager import CatalogDataManager
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.units import unit_instance_factory

import matplotlib
matplotlib.use('TkAgg')
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg  # noqa: E402
from matplotlib.figure import Figure  # noqa: E402


class CatalogDataPlotForm(Toplevel):
    """
    This form is where we display the processed catalog data to the user to allow them to see variation in each
    parameter.
    """
    class ExitCode(Enum):
        OK = auto()
        CANCEL = auto()

    def __init__(self, parent_window, cdm: CatalogDataManager, eq: BaseEquipment):
        super().__init__(parent_window)
        self.exit_code = CatalogDataPlotForm.ExitCode.CANCEL
        p = 4
        Label(self, text="Here's some info about this catalog data\nClick through the plots to inspect").pack(
            side=TOP, expand=True, fill=X
        )
        style = Style(self)
        style.configure('my.TNotebook', tabposition='wn')
        plot_notebook = Notebook(self, style='my.TNotebook')
        plot_data = []
        for col_num in range(len(eq.headers().name_array())):
            line_title = eq.headers().name_array()[col_num]
            line_unit_type = eq.headers().unit_array()[col_num]
            dummy_unit_instance = unit_instance_factory(0.0, line_unit_type)
            line_unit_string = dummy_unit_instance.get_unit_string_map()[dummy_unit_instance.calculation_unit_id()]
            data_vector = [_row[col_num] for _row in cdm.final_data_matrix]
            # could allow x_values to be passed in if we want a specific plot shape
            x_values = list(range(len(data_vector)))
            plot_data.append(
                (line_title, x_values, data_vector, line_unit_string)
            )
        for pd in plot_data:
            plot_frame = Frame(plot_notebook)
            plot_frame.pack(side=TOP, expand=True, fill=BOTH)
            plot_notebook.add(plot_frame, text=pd[0])
            fig = Figure(figsize=(7, 5))
            a = fig.add_subplot(111)
            a.plot(np.array(pd[1]), np.array(pd[2]))
            a.set_title("Catalog Data Display", fontsize=16)
            a.set_ylabel(f"[{pd[3]}]", fontsize=14)
            a.set_xlabel("Catalog Data Points (no order)", fontsize=14)
            canvas = FigureCanvasTkAgg(fig, master=plot_frame)
            canvas.get_tk_widget().pack()
            canvas.draw()
        plot_notebook.pack(side=TOP, expand=True, fill=BOTH, padx=p, pady=p)
        button_frame = Frame(self)
        Button(button_frame, text=u"\U0001f44D Looks good, generate parameters", command=self.ok).grid(
            row=0, column=0, padx=p, pady=p
        )
        Button(button_frame, text="\U0001f44E Something is wrong -- Cancel Wizard", command=self.cancel).grid(
            row=0, column=1, padx=p, pady=p
        )
        button_frame.grid_columnconfigure(ALL, weight=1)
        button_frame.pack(side=TOP, expand=False, fill=X)
        self.grab_set()
        self.transient(parent_window)

    def ok(self):
        self.exit_code = CatalogDataPlotForm.ExitCode.OK
        self.finish()

    def cancel(self):
        self.exit_code = CatalogDataPlotForm.ExitCode.CANCEL
        self.finish()

    def finish(self):
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from random import randint
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_cooling_curve import WaterToAirHeatPumpCoolingCurveFit

    window = Tk()
    _cdm = CatalogDataManager()
    d = []
    for i in range(3):
        row = []
        for j in range(8):
            row.append(randint(1, 10))
        d.append(row)
    _cdm.add_base_data(d)
    _cdm.apply_correction_factors(3, -1, -1)
    start = CatalogDataPlotForm(window, _cdm, WaterToAirHeatPumpCoolingCurveFit())
    window.mainloop()



from json import dumps
from typing import Callable, List, Tuple

from energyplus_pet.equipment.common_curves import CommonCurves
from energyplus_pet.data_manager import CatalogDataManager
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.equipment.equip_types import EquipType
from energyplus_pet.equipment.column_header import ColumnHeaderArray, ColumnHeader
from energyplus_pet.exceptions import EnergyPlusPetException
from energyplus_pet.units import UnitType


class WaterToAirHeatPumpHeatingCurveFit(BaseEquipment):
    """
    This class represents a water-to-air heat pump, heating mode, using the model formulation:
    Q = Q_rated * (A + B * Tl + C * Ts + D * Vl + E * Vs)
    P = P_rated * (F + G * Tl + H * Ts + I * Vl + J * Vs)
    where:
    Q is the load side heat transfer rate
    P is the heating power
    _rated indicates the nominal operating value
    A-J are curve fit coefficients
    Tl and Ts are scaled load and source side inlet temps (T/283.15) where original T is in Kelvin
    Vl and Vs are scaled load and source side volume flow rates (V/V_rated)
    """

    def __init__(self):
        # need some rated parameters that we get from the user for scaling, reporting, etc.
        self.rated_air_volume_flow_key = 'vl'
        self.rated_air_volume_flow = 0.0
        self.rated_water_volume_flow_key = 'vs'
        self.rated_water_volume_flow = 0.0
        self.rated_heating_capacity_key = 'qh'
        self.rated_heating_capacity = 0.0
        self.rated_heating_power_key = 'cp'
        self.rated_heating_power = 0.0
        # store some individual arrays for each of the dependent variable input columns
        self.catalog_heating_capacity = []
        self.catalog_heating_power = []
        # these eventually become the actual parameter arrays
        self.heating_capacity_params = []
        self.heating_power_params = []
        # these represent a metric for the quality of the regression
        self.heating_capacity_avg_err = 0.0
        self.heating_power_avg_err = 0.0
        # store the predicted outputs that are calculated from the generated parameters
        self.predicted_heating_capacity = []
        self.predicted_heating_power = []
        self.percent_error_heating_capacity = []
        self.percent_error_heating_power = []
        # store the headers on the instance, so we don't reconstruct it every call to headers()
        self._headers = ColumnHeaderArray(
            [
                ColumnHeader("Water-side Entering Temp", UnitType.Temperature),
                ColumnHeader("Water-side Volume Flow", UnitType.Flow),
                ColumnHeader("Air-side Entering Dry-bulb Temp", UnitType.Temperature, db=True),
                ColumnHeader("Air-side Volume Flow", UnitType.Flow),
                ColumnHeader("Heating Capacity", UnitType.Power),
                ColumnHeader("Heating Power", UnitType.Power)
            ]
        )

    def this_type(self) -> EquipType:
        return EquipType.WAHP_Heating_CurveFit

    def name(self) -> str:
        return "Water to Air Heat Pump, Heating Coil, Curve Fit Formulation"

    def short_name(self) -> str:
        return "WAHP-Heating-CurveFit"

    def get_required_constant_parameters(self) -> List[BaseEquipment.RequiredConstantParameter]:
        return [
            # need some rated parameters that we get from the user for scaling, reporting, etc.
            BaseEquipment.RequiredConstantParameter(
                self.rated_air_volume_flow_key,
                "Rated Air Flow Rate",
                "This is a nominal flow rate value for the air-side of the coil",
                UnitType.Flow,
                0.0006887,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_water_volume_flow_key,
                "Rated Water Flow Rate",
                "This is a nominal flow rate value for the water-side of the coil",
                UnitType.Flow,
                0.0001892,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_heating_capacity_key,
                "Rated Total Heating Capacity",
                "This is a nominal value of the load-side heating capacity of the coil",
                UnitType.Power,
                3.513,
            ),
            BaseEquipment.RequiredConstantParameter(
                self.rated_heating_power_key,
                "Rated Heating Power",
                "This is a nominal value of the heating power for this coil",
                UnitType.Power,
                0.900,
            )
        ]

    def set_required_constant_parameter(self, parameter_id: str, new_value: float) -> None:
        if parameter_id == self.rated_air_volume_flow_key:
            self.rated_air_volume_flow = new_value
        elif parameter_id == self.rated_water_volume_flow_key:
            self.rated_water_volume_flow = new_value
        elif parameter_id == self.rated_heating_capacity_key:
            self.rated_heating_capacity = new_value
        elif parameter_id == self.rated_heating_power_key:
            self.rated_heating_power = new_value
        else:
            raise EnergyPlusPetException("Bad parameter ID in set_required_constant_parameter")

    def headers(self) -> ColumnHeaderArray:
        return self._headers

    def to_eplus_idf_object(self) -> str:
        object_name = "Coil:Heating:WaterToAirHeatPump:EquationFit"
        fields = [
            ("Name", 'Your Coil Name'),
            ("Water Inlet Node Name", 'Your Coil Source Side Inlet Node'),
            ("Water Outlet Node Name", 'Your Coil Source Side Outlet Node'),
            ("Air Inlet Node Name", 'Your Coil Load Side Inlet Node'),
            ("Air Outlet Node Name", 'Your Coil Load Side Outlet Node'),
            ("Rated Air Flow Rate", self.rated_air_volume_flow),
            ("Rated Water Flow Rate", self.rated_water_volume_flow),
            ("Gross Rated Heating Capacity", self.rated_heating_capacity),
            ("Gross Rated Heating COP", round(self.rated_heating_capacity / self.rated_heating_power, 4)),
            ("Rated Entering Water Temperature", ''),
            ("Rated Entering Air Dry-Bulb Temp", ''),
            ("Ratio of Rated Heating Capacity to Rated Cooling Capacity", ''),
            ("Heating Capacity Curve Name", 'TotalCapacityCurve'),
            ("Heating Power Consumption Curve Name", 'HeatingPowerCurve'),
        ]
        coil_object_string = self.fill_eplus_object_format(object_name, fields)

        reused_limits = [
            ("Minimum Value of w", -100), ("Maximum Value of w", 100),
            ("Minimum Value of x", -100), ("Maximum Value of x", 100),
            ("Minimum Value of y", -100), ("Maximum Value of y", 100),
            ("Minimum Value of z", -100), ("Maximum Value of z", 100),
        ]

        object_name = "Curve:QuadLinear"
        fields = [
            ("Name", "TotalCapacityCurve"),
            *[(f"Coefficient{i}", self.heating_capacity_params[i]) for i in range(5)],
            *reused_limits
        ]
        total_curve_output = self.fill_eplus_object_format(object_name, fields)

        object_name = "Curve:QuadLinear"
        fields = [
            ("Name", "HeatingPowerCurve"),
            *[(f"Coefficient{i}", self.heating_power_params[i]) for i in range(5)],
            *reused_limits
        ]
        sensible_curve_output = self.fill_eplus_object_format(object_name, fields)

        return '\n'.join([
            BaseEquipment.current_eplus_version_object_idf(),
            coil_object_string,
            total_curve_output, sensible_curve_output
        ])

    def to_parameter_summary(self) -> str:
        output = f"""{self.name()}
**Begin Nomenclature**
HC: Heating Capacity
HP: Heating Power Consumption
TLI: Entering Load-side Temperature
TSI: Entering Source-side Temperature
VLI: Entering Load-side Flow Rate
VSI: Entering Source-side Flow Rate
Subscript _R: Rated Value
Subscript _#: Coefficient #
**End Nomenclature**

**Begin Governing Equations**
(HC/HC_R) = HC_1 + HC_2*(TLI/TLI_R) + HC_3*(TSI/TSI_R) + HC_4*(VLI/VLI_R) + HC_5*(VSI/VSI_R)
(HP/HP_R) = HP_1 + HP_2*(TLI/TLI_R) + HP_3*(TSI/TSI_R) + HP_4*(VLI/VLI_R) + HP_5*(VSI/VSI_R)
**End Governing Equations**

**Begin Reporting Parameters**
Rated Load-side Heating Capacity: {self.rated_heating_capacity}
Rated Heating Power Consumption: {self.rated_heating_power}
Rated Load-side Volumetric Flow Rate: {self.rated_air_volume_flow}
Rated Source-side Volumetric Flow Rate: {self.rated_water_volume_flow}
"""
        for i, c in enumerate(self.heating_capacity_params):
            output += f"Heating Capacity Coefficient HC_{i + 1}: {round(c, 4)}\n"
        for i, c in enumerate(self.heating_power_params):
            output += f"Heating Power Consumption Coefficient HC_{i + 1}: {round(c, 4)}\n"
        output += "**End Reporting Parameters**"
        return output

    def to_eplus_epjson_object(self) -> str:
        coil_object = {'Your Coil Name': {
            'water_inlet_node_name': 'Your Coil Source Side Inlet Node',
            'water_outlet_node_name': 'Your Coil Source Side Outlet Node',
            'air_inlet_node_name': 'Your Coil Load Side Inlet Node',
            'air_outlet_node_name': 'Your Coil Load Side Outlet Node',
            'rated_air_flow_rate': self.rated_air_volume_flow,
            'rated_water_flow_rate': self.rated_water_volume_flow,
            'gross_rated_heating_capacity': self.rated_heating_capacity,
            'gross_rated_heating_cop': self.rated_heating_capacity / self.rated_heating_power,
            'rated_entering_water_temperature': '',
            'rated_entering_air_dry_bulb_temperature': '',
            'ratio_of_rated_heating_capacity_to_rated_cooling_capacity': '',
            'heating_capacity_curve_name': 'TotalCapacityCurve',
            'heating_power_consumption_curve_name': 'HeatingPowerCurve',
        }}
        reused_limits = {
            "minimum_value_of_w": -100, "maximum_value_of_w": 100,
            "minimum_value_of_x": -100, "maximum_value_of_x": 100,
            "minimum_value_of_y": -100, "maximum_value_of_y": 100,
            "minimum_value_of_z": -100, "maximum_value_of_z": 100,
        }
        curves = {'TotalCapacityCurve': {
            "coefficient1_constant": self.heating_capacity_params[0],
            "coefficient2_w": self.heating_capacity_params[1],
            "coefficient3_x": self.heating_capacity_params[2],
            "coefficient4_y": self.heating_capacity_params[3],
            "coefficient5_z": self.heating_capacity_params[4],
            **reused_limits
        }, 'HeatingPowerCurve': {
            "coefficient1_constant": self.heating_power_params[0],
            "coefficient2_w": self.heating_power_params[1],
            "coefficient3_x": self.heating_power_params[2],
            "coefficient4_y": self.heating_power_params[3],
            "coefficient5_z": self.heating_power_params[4],
            **reused_limits
        }}

        epjson_object = {
            **BaseEquipment.current_eplus_version_object_epjson(),
            'Coil:Heating:WaterToAirHeatPump:EquationFit': coil_object,
            'Curve:QuadLinear': curves
        }
        return dumps(epjson_object, indent=2)

    def get_number_of_progress_steps(self) -> int:
        return 3  # read data, hc curve fit, power curve fit

    def minimum_data_points_for_generation(self) -> int:
        return 5

    def generate_parameters(
            self, data_manager: CatalogDataManager, cb_progress_increment: Callable, cb_progress_done: Callable
    ):
        # step 1, read the data into arrays (will be used for both scaling calculations and plotting later)
        # these should already be set to zero, but not sure if the main form will reinitialize the equip-instance or not
        self.catalog_heating_capacity = [x[4] for x in data_manager.final_data_matrix]
        self.catalog_heating_power = [x[5] for x in data_manager.final_data_matrix]
        scaled_water_inlet_temp = []
        scaled_water_flow_rate = []
        scaled_air_inlet_temp = []
        scaled_air_flow_rate = []
        scaled_heating_capacity = []
        scaled_heating_power = []
        for data_row in data_manager.final_data_matrix:
            scaled_water_inlet_temp.append((data_row[0] + 273.15) / (10.0 + 273.15))  # T_ref defined by HP model
            scaled_water_flow_rate.append(data_row[1] / self.rated_water_volume_flow)
            scaled_air_inlet_temp.append((data_row[2] + 273.15) / (10.0 + 273.15))
            scaled_air_flow_rate.append(data_row[3] / self.rated_air_volume_flow)
            scaled_heating_capacity.append(data_row[4] / self.rated_heating_capacity)
            scaled_heating_power.append(data_row[5] / self.rated_heating_power)
        cb_progress_increment()

        independent_var_arrays = (
            scaled_air_inlet_temp,
            scaled_water_inlet_temp,
            scaled_air_flow_rate,
            scaled_water_flow_rate
        )

        self.heating_capacity_params, self.heating_capacity_avg_err = self.do_one_curve_fit(
            CommonCurves.heat_pump_5_coefficient_curve,
            independent_var_arrays,
            scaled_heating_capacity
        )
        cb_progress_increment()

        self.heating_power_params, self.heating_power_avg_err = self.do_one_curve_fit(
            CommonCurves.heat_pump_5_coefficient_curve,
            independent_var_arrays,
            scaled_heating_power
        )
        cb_progress_increment()

        # now just recalculate the values at each catalog data point
        self.predicted_heating_capacity, self.percent_error_heating_capacity = self.eval_curve_at_points(
            lambda *x: CommonCurves.heat_pump_5_coefficient_curve_raw_value(*x, self.rated_heating_capacity),
            independent_var_arrays,
            self.heating_capacity_params,
            self.catalog_heating_capacity
        )
        self.predicted_heating_power, self.percent_error_heating_power = self.eval_curve_at_points(
            lambda *x: CommonCurves.heat_pump_5_coefficient_curve_raw_value(*x, self.rated_heating_power),
            independent_var_arrays,
            self.heating_power_params,
            self.catalog_heating_power
        )
        cb_progress_done(True)

    def get_absolute_plot_data(self) -> Tuple:
        return (
            ('Total Heat Transfer Model', 'line', 'red', self.predicted_heating_capacity),
            ('Total Heat Transfer Catalog', 'point', 'red', self.catalog_heating_capacity),
            ('Heating Power Model', 'line', 'green', self.predicted_heating_power),
            ('Heating Power Catalog', 'point', 'green', self.catalog_heating_power),
        )

    def get_error_plot_data(self) -> Tuple:
        return (
            ('Total Heat Transfer % Error', 'line', 'red', self.percent_error_heating_capacity),
            ('Heating Power % Error', 'line', 'green', self.percent_error_heating_power),
        )

    def get_extra_regression_metrics(self) -> Tuple:
        return (
            (
                "Total Heat Transfer Average curve-fit error (1 standard deviation)",
                self.heating_capacity_avg_err
            ),
            (
                "Heating Power Average curve-fit error (1 standard deviation)",
                self.heating_power_avg_err
            )
        )



from enum import auto, Enum
from json import dumps
from typing import Dict, List


class CorrectionFactorType(Enum):
    Multiplier = auto()
    Replacement = auto()
    CombinedDbWb = auto()


class CorrectionFactor:
    """
    This class represents a small data set that is used to modify the base data and create a larger final data set.
    Often times manufacturers will provide equipment data with some independent variables held constant.
    From this data alone, it is not possible to generate a curve fit that is a function of that variable.
    It is common for the manufacturers to provide small separate tables of correction factor data that describe how
    the dependent variables behave as those originally constant values change, using multipliers for the dependent data.
    """
    def __init__(self, name: str):
        """
        Create a new CorrectionFactor instance.  From a user perspective, the correction factor is generated in two
        parts: inputting basic summary data, then inputting detailed tabular data.  Behind the scenes, the construction
        of a correction factor starts with just a name passed in here.  After that, there is a set of summary data
        members initialized to meaningful values for the widget to show.  Finally, once details have been read from
        the user, the data is set here.
        :param name:
        """
        self.name: str = name
        self.check_ok_messages: List[str] = []

        # The following are the variables that define this correction factor summary, initialize them as needed and the
        # widget should reflect the initialized values by setting Tk Variables appropriately.
        self.num_corrections: int = 5
        self.correction_type: CorrectionFactorType = CorrectionFactorType.Multiplier
        self.base_column_index: int = -1
        self._columns_to_modify: List[int] = []

        # for normal corrections, just keep one array of base column data
        self.base_correction: List[float] = []  # could be multipliers or replacement values
        # for db/wb corrections, just keep two separate arrays
        self.base_correction_db: List[float] = []  # could be multipliers or replacement values
        self.base_correction_wb: List[float] = []  # could be multipliers or replacement values
        # keep a dict of dependent variable modifiers keyed off of the equipment column number
        self.mod_correction_data_column_map: Dict[int, List[float]] = {}

    @property
    def columns_to_modify(self) -> List[int]:
        """Returns the list of columns that this correction factor is designed to modify"""
        return self._columns_to_modify

    @columns_to_modify.setter
    def columns_to_modify(self, mod_column_indices: List[int]):
        """Sets the list of columns that this correction factor modifies, initializes data column map internally"""
        self._columns_to_modify = mod_column_indices
        for i in mod_column_indices:
            self.mod_correction_data_column_map[i] = []

    def describe(self) -> str:
        """Returns a multiline descriptive string for this correction factor"""
        response = f"CorrectionFactor: {self.name}\n"
        response += f"* {self.num_corrections} {self.correction_type.name} rows\n"
        if self.correction_type != CorrectionFactorType.CombinedDbWb:
            response += f"* base column: {self.base_column_index}\n"
            response += f"* Base Correction Array: {self.base_correction}\n"
        else:
            response += "* base columns are equipment db/wb columns\n"
            response += f"* Base dry-bulb correction array: {self.base_correction_db}"
            response += f"* Base wet-bulb correction array: {self.base_correction_wb}"
        response += f"* modifies columns {self._columns_to_modify}\n"
        response += "* Mod Correction Matrix:\n"
        response += dumps(self.mod_correction_data_column_map, indent=2)
        return response

    def check_ok(self, db_column: int, wb_column: int, summary_only: bool = False) -> bool:
        """
        Checks values of this correction factor and returns True or False to indicate success.
        If False, the client can check the .check_ok_last_message for a string message about what is wrong

        :param db_column: The dry-bulb column from the current equipment headers().get_db_column()
        :param wb_column: The wet-bulb column from the current equipment headers().get_wb_column()
        :param summary_only: If this is True, it will only check summary data, not the full data set
        :return: True if everything looks good, False if not
        """
        self.check_ok_messages.clear()
        if self.num_corrections < 1:
            self.check_ok_messages.append(
                f"# of corrections ({self.num_corrections}) is less than 1, this is invalid."
            )
        if self.correction_type not in list(CorrectionFactorType):
            self.check_ok_messages.append(
                f"Correction factor type appears invalid: ({self.correction_type})."
            )
        if not self.correction_type == CorrectionFactorType.CombinedDbWb:
            if self.base_column_index < 0:
                self.check_ok_messages.append(
                    f"Base column index ({self.base_column_index}) is less than 0, invalid."
                )
            if self.base_column_index in self._columns_to_modify:
                self.check_ok_messages.append(
                    "Base column index appears in modification column list, invalid."
                )
        else:  # is a wet-bulb/dry-bulb type
            if db_column in self._columns_to_modify:
                self.check_ok_messages.append(
                    "Dry-bulb column index appears in modification column list, invalid."
                )
            if wb_column in self._columns_to_modify:
                self.check_ok_messages.append(
                    "Wet-bulb column index appears in modification column list, invalid."
                )
        if not summary_only:
            if not self.correction_type == CorrectionFactorType.CombinedDbWb:
                if len(self.base_correction) != self.num_corrections:
                    self.check_ok_messages.append(
                        'Size of base corrections does not match num_corrections, invalid.'
                    )
            else:
                if len(self.base_correction_db) != self.num_corrections:
                    self.check_ok_messages.append(
                        'Size of base dry-bulb corrections does not match num_corrections, invalid.'
                    )
                if len(self.base_correction_wb) != self.num_corrections:
                    self.check_ok_messages.append(
                        'Size of base wet-bulb corrections does not match num_corrections, invalid.'
                    )
            if len(self.mod_correction_data_column_map) != len(self._columns_to_modify):
                self.check_ok_messages.append(
                    'Size of mod correction data does not match columns_to_modify, invalid.'
                )
            for c in self._columns_to_modify:
                if c not in self.mod_correction_data_column_map:
                    self.check_ok_messages.append(
                        f"Did not find mod column {c} in mod data map keys."
                    )
                elif len(self.mod_correction_data_column_map[c]) != self.num_corrections:
                    self.check_ok_messages.append(
                        f"Size of column {c} data array in mode data map does not match num_corrections = "
                        f"{self.num_corrections}, invalid. "
                    )
        if self.check_ok_messages:  # the message has content, must be a problem
            return False
        return True



from abc import abstractmethod
from collections import OrderedDict
from enum import Enum, auto
from typing import List, Type, OrderedDict as TypingOD

from energyplus_pet.exceptions import EnergyPlusPetException


class UnitType(Enum):
    Power = auto()
    Flow = auto()
    Temperature = auto()
    Dimensionless = auto()
    Pressure = auto()
    Length = auto()
    RotationalSpeed = auto()


class BaseValueWithUnit:
    """
    This class represents an abstract floating point value with units attached.
    In this library, if a component is trying to define an interface, for example, defining a header column, it should
    just use the UnitType enumeration, because there is no value associated with that column.
    This class is mostly just used while tabular forms are open to handle users entering data in various units and
    normalizing them before sending the properly formed data to the equipment processing routines.
    """

    def __init__(
            self, initial_value: float, name: str = "<Unnamed>", description: str = '<...>', initial_unit_id: str = None
    ):
        self.value = initial_value
        self.name = name
        self.description = description
        if initial_unit_id is not None:
            if initial_unit_id not in self.get_unit_ids():
                raise EnergyPlusPetException(
                    f"Invalid unit ID in unit constructor for class {self.__class__.__name__} = {initial_unit_id}"
                )
            self.units = initial_unit_id
        else:
            self.units = self.calculation_unit_id()

    @abstractmethod
    def get_unit_type(self) -> UnitType:  # pragma: no cover
        pass

    @staticmethod
    @abstractmethod
    def get_unit_ids() -> List[str]:  # pragma: no cover
        pass

    @staticmethod
    @abstractmethod
    def get_unit_string_map() -> TypingOD[str, str]:  # pragma: no cover
        pass

    @classmethod
    def get_id_from_unit_string(cls, current_units_string: str) -> str:
        """
        Looks up the ID of a unit for the current type given the unit string

        :param current_units_string: A string of units to search for, such as 'kg/s'
        :return: An internal string ID of the unit matching what comes from get_unit_ids()
        """
        for k, v in cls.get_unit_string_map().items():
            if v == current_units_string:
                return k
        raise EnergyPlusPetException(
            f"No ID for unit string: {current_units_string}; possible units = {cls.get_unit_string_map().values()}"
        )

    @staticmethod
    @abstractmethod
    def calculation_unit_id() -> str:  # pragma: no cover
        pass

    @staticmethod
    @abstractmethod
    def base_ip_unit_id() -> str:  # pragma: no cover
        pass

    @staticmethod
    @abstractmethod
    def base_si_unit() -> str:  # pragma: no cover
        pass

    @abstractmethod
    def convert_to_calculation_unit(self):  # pragma: no cover
        pass

    def __str__(self) -> str:
        return f"{self.value} [{self.get_unit_string_map()[self.units]}]"


class DimensionlessValue(BaseValueWithUnit):
    Dimensionless = __qualname__ + "-"

    def get_unit_type(self) -> UnitType:
        return UnitType.Dimensionless

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [DimensionlessValue.Dimensionless]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[DimensionlessValue.Dimensionless] = "--"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return DimensionlessValue.Dimensionless

    @staticmethod
    def base_ip_unit_id() -> str:
        return DimensionlessValue.Dimensionless

    @staticmethod
    def base_si_unit() -> str:
        return DimensionlessValue.Dimensionless

    def convert_to_calculation_unit(self):
        return


class PowerValue(BaseValueWithUnit):
    Kilowatts = __qualname__ + "kW"
    Watts = __qualname__ + "W"
    BTU_hour = __qualname__ + "BtuH"
    MBTU_hour = __qualname__ + "MBtuH"

    def get_unit_type(self) -> UnitType:
        return UnitType.Power

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [PowerValue.Watts, PowerValue.Kilowatts, PowerValue.BTU_hour, PowerValue.MBTU_hour]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[PowerValue.Kilowatts] = "kW"
        od[PowerValue.Watts] = "W"
        od[PowerValue.BTU_hour] = "Btu/hr"
        od[PowerValue.MBTU_hour] = "MBtu/hr"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return PowerValue.Kilowatts

    @staticmethod
    def base_ip_unit_id() -> str:
        return PowerValue.MBTU_hour

    @staticmethod
    def base_si_unit() -> str:
        return PowerValue.Kilowatts

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == PowerValue.Watts:
            self.value /= 1000.0
        elif self.units == PowerValue.BTU_hour:
            factor = 1 / (3.412 * 1000)
            self.value *= factor
        elif self.units == PowerValue.MBTU_hour:
            factor = 1000 / (3.412 * 1000)
            self.value *= factor
        self.units = self.calculation_unit_id()


class FlowValue(BaseValueWithUnit):
    GPM = __qualname__ + "GPM"
    CFM = __qualname__ + "CFM"
    M3S = __qualname__ + "M3S"

    def get_unit_type(self) -> UnitType:
        return UnitType.Flow

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [FlowValue.GPM, FlowValue.CFM, FlowValue.M3S]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[FlowValue.GPM] = "GPM"
        od[FlowValue.CFM] = "CFM"
        od[FlowValue.M3S] = "m^3/s"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return FlowValue.M3S

    @staticmethod
    def base_ip_unit_id() -> str:
        return FlowValue.GPM

    @staticmethod
    def base_si_unit() -> str:
        return FlowValue.M3S

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == FlowValue.CFM:
            self.value *= 0.0004719474432
        elif self.units == FlowValue.GPM:
            self.value *= 0.00006309
        self.units = self.calculation_unit_id()


class TemperatureValue(BaseValueWithUnit):
    F = __qualname__ + "F"
    C = __qualname__ + "C"
    K = __qualname__ + "K"

    def get_unit_type(self) -> UnitType:
        return UnitType.Temperature

    @staticmethod
    def get_unit_ids() -> List[int]:
        return [TemperatureValue.F, TemperatureValue.C, TemperatureValue.K]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[TemperatureValue.F] = "deg F"
        od[TemperatureValue.C] = "deg C"
        od[TemperatureValue.K] = "Kelvin"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return TemperatureValue.C

    @staticmethod
    def base_ip_unit_id() -> str:
        return TemperatureValue.F

    @staticmethod
    def base_si_unit() -> str:
        return TemperatureValue.C

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == TemperatureValue.F:
            self.value = (self.value - 32.0) / 1.8
        elif self.units == TemperatureValue.K:
            self.value -= 273.15
        self.units = self.calculation_unit_id()


class PressureValue(BaseValueWithUnit):
    Pa = __qualname__ + "Pa"
    KPa = __qualname__ + "kPa"
    Atm = __qualname__ + "Atm"
    PSI = __qualname__ + "PSI"

    def get_unit_type(self) -> UnitType:
        return UnitType.Pressure

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [PressureValue.Pa, PressureValue.KPa, PressureValue.Atm, PressureValue.PSI]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[PressureValue.Pa] = "Pa"
        od[PressureValue.KPa] = "kPa"
        od[PressureValue.Atm] = "atm"
        od[PressureValue.PSI] = "psi"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return PressureValue.Pa

    @staticmethod
    def base_ip_unit_id() -> str:
        return PressureValue.PSI

    @staticmethod
    def base_si_unit() -> str:
        return PressureValue.Pa

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == PressureValue.KPa:
            self.value *= 1000.0
        elif self.units == PressureValue.Atm:
            self.value *= 101325.0
        elif self.units == PressureValue.PSI:
            self.value *= 6894.757
        self.units = self.calculation_unit_id()


class LengthValue(BaseValueWithUnit):
    Meters = __qualname__ + "M"
    Feet = __qualname__ + "Ft"
    Inches = __qualname__ + "In"
    Centimeters = __qualname__ + "CM"
    Millimeters = __qualname__ + "MM"

    def get_unit_type(self) -> UnitType:
        return UnitType.Length

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [
            LengthValue.Meters, LengthValue.Feet, LengthValue.Inches, LengthValue.Centimeters, LengthValue.Millimeters
        ]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[LengthValue.Meters] = "Meters"
        od[LengthValue.Feet] = "Feet"
        od[LengthValue.Inches] = "Inches"
        od[LengthValue.Centimeters] = "Centimeters"
        od[LengthValue.Millimeters] = "Millimeters"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return LengthValue.Meters

    @staticmethod
    def base_ip_unit_id() -> str:
        return LengthValue.Inches

    @staticmethod
    def base_si_unit() -> str:
        return LengthValue.Meters

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == LengthValue.Feet:
            self.value *= 0.3048
        elif self.units == LengthValue.Inches:
            self.value *= 0.0254
        elif self.units == LengthValue.Centimeters:
            self.value *= 0.01
        elif self.units == LengthValue.Millimeters:
            self.value *= 0.001
        self.units = self.calculation_unit_id()


class RotationSpeedValue(BaseValueWithUnit):
    RevsPerSecond = __qualname__ + "Rps"
    RevsPerMinute = __qualname__ + "Rpm"
    RadiansPerSecond = __qualname__ + "RadPs"

    def get_unit_type(self) -> UnitType:
        return UnitType.RotationalSpeed

    @staticmethod
    def get_unit_ids() -> List[str]:
        return [RotationSpeedValue.RevsPerSecond, RotationSpeedValue.RevsPerMinute, RotationSpeedValue.RadiansPerSecond]

    @staticmethod
    def get_unit_string_map() -> TypingOD[str, str]:
        od = OrderedDict()
        od[RotationSpeedValue.RevsPerSecond] = "Revs/Sec"
        od[RotationSpeedValue.RevsPerMinute] = "Revs/Min"
        od[RotationSpeedValue.RadiansPerSecond] = "Rads/Sec"
        return od

    @staticmethod
    def calculation_unit_id() -> str:
        return RotationSpeedValue.RevsPerSecond

    @staticmethod
    def base_ip_unit_id() -> str:
        return RotationSpeedValue.RevsPerMinute

    @staticmethod
    def base_si_unit() -> str:
        return RotationSpeedValue.RevsPerSecond

    def convert_to_calculation_unit(self):
        if self.units == self.calculation_unit_id():
            return
        elif self.units == RotationSpeedValue.RadiansPerSecond:
            self.value /= 6.2831853
        elif self.units == RotationSpeedValue.RevsPerMinute:
            self.value /= 60.0
        self.units = self.calculation_unit_id()


def unit_class_factory(unit_type: UnitType) -> Type[BaseValueWithUnit]:
    if unit_type == UnitType.Power:
        return PowerValue
    elif unit_type == UnitType.Flow:
        return FlowValue
    elif unit_type == UnitType.Temperature:
        return TemperatureValue
    elif unit_type == UnitType.Dimensionless:
        return DimensionlessValue
    elif unit_type == UnitType.Pressure:
        return PressureValue
    elif unit_type == UnitType.Length:
        return LengthValue
    elif unit_type == UnitType.RotationalSpeed:
        return RotationSpeedValue
    else:
        raise EnergyPlusPetException(f"Bad unit_type input sent to unit_class_factory: \"{str(unit_type)}\", aborting.")


def unit_instance_factory(value: float, unit_type: UnitType) -> BaseValueWithUnit:
    return unit_class_factory(unit_type)(value)



from enum import Enum, auto


class EquipType(Enum):
    """Enumeration class of all known equipment types"""
    InvalidType = auto()  # used to return an invalid selection wherever needed
    WAHP_Heating_CurveFit = auto()
    WAHP_Heating_PE = auto()
    WAHP_Cooling_CurveFit = auto()
    WAHP_Cooling_PE = auto()
    WWHP_Heating_CurveFit = auto()
    WWHP_Cooling_CurveFit = auto()
    Pump_ConstSpeed_ND = auto()


class EquipTypeUniqueStrings:
    InvalidType = 'InvalidType'
    WAHP_Heating_CurveFit = 'WAHP_Heating_CurveFit'
    WAHP_Heating_PE = 'WAHP_Heating_PE'
    WAHP_Cooling_CurveFit = 'WAHP_Cooling_CurveFit'
    WAHP_Cooling_PE = 'WAHP_Cooling_PE'
    WWHP_Heating_CurveFit = 'WWHP_Heating_CurveFit'
    WWHP_Cooling_CurveFit = 'WWHP_Cooling_CurveFit'
    Pump_ConstSpeed_ND = 'Pump_ConstSpeed_ND'

    @staticmethod
    def get_equip_type_from_unique_string(equip_type_string) -> EquipType:
        if equip_type_string == EquipTypeUniqueStrings.InvalidType:
            return EquipType.InvalidType
        elif equip_type_string == EquipTypeUniqueStrings.WAHP_Heating_CurveFit:
            return EquipType.WAHP_Heating_CurveFit
        elif equip_type_string == EquipTypeUniqueStrings.WAHP_Heating_PE:
            return EquipType.WAHP_Heating_PE
        elif equip_type_string == EquipTypeUniqueStrings.WAHP_Cooling_CurveFit:
            return EquipType.WAHP_Cooling_CurveFit
        elif equip_type_string == EquipTypeUniqueStrings.WAHP_Cooling_PE:
            return EquipType.WAHP_Cooling_PE
        elif equip_type_string == EquipTypeUniqueStrings.WWHP_Heating_CurveFit:
            return EquipType.WWHP_Heating_CurveFit
        elif equip_type_string == EquipTypeUniqueStrings.WWHP_Cooling_CurveFit:
            return EquipType.WWHP_Cooling_CurveFit
        elif equip_type_string == EquipTypeUniqueStrings.Pump_ConstSpeed_ND:
            return EquipType.Pump_ConstSpeed_ND



from tkinter import Toplevel, Button, Label, TOP, X, CENTER, HORIZONTAL, LEFT
from tkinter.ttk import Separator


class PetMessageForm(Toplevel):
    """Tkinter dialogs look a little funny on Linux, just building a basic form myself"""
    def __init__(self, parent_window, title: str, message: str, subtitle: str = '', justify_message_left: bool = False):
        super().__init__(parent_window)
        self.title(title)
        if subtitle:
            Label(self, text=subtitle).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
            Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        justify = LEFT if justify_message_left else CENTER
        Label(self, text=message, justify=justify).pack(side=TOP, fill=X, expand=True, padx=3, pady=3)
        Button(self, text="OK", command=self._done).pack(side=TOP, expand=False, anchor=CENTER, padx=3, pady=3)
        self.grab_set()
        self.transient(parent_window)

    def _done(self):
        self.grab_release()
        self.destroy()
